// This file contains FFT-related kernels:
// * Conversion between Speex format and Apple OpenCL FFT format
// * Hardcoded FFT kernels based on Apple OpenCL FFT (to avoid library overhead)
// * Various kernels where the FFT is combined with other operations
#pragma OPENCL EXTENSION cl_khr_byte_addressable_store : enable 
#ifndef M_PI
#define M_PI 0x1.921fb54442d18p+1
#endif
float2 complexMul(float2 a,float2 B) { return (float2)(mad(-(a).y, (B).y, (a).x * (B).x), mad((a).y, (B).x, (a).x * (B).y));}
#define conj(a) ((float2)((a).x, -(a).y))
#define conjTransp(a) ((float2)(-(a).y, (a).x))

#define fftKernel2(a,dir) \
{ \
    float2 c = (a)[0];    \
    (a)[0] = c + (a)[1];  \
    (a)[1] = c - (a)[1];  \
}

#define fftKernel2S(d1,d2,dir) \
{ \
    float2 c = (d1);   \
    (d1) = c + (d2);   \
    (d2) = c - (d2);   \
}

#define fftKernel4(a,dir) \
{ \
    fftKernel2S((a)[0], (a)[2], dir); \
    fftKernel2S((a)[1], (a)[3], dir); \
    fftKernel2S((a)[0], (a)[1], dir); \
    (a)[3] = (float2)(dir)*(conjTransp((a)[3])); \
    fftKernel2S((a)[2], (a)[3], dir); \
    float2 c = (a)[1]; \
    (a)[1] = (a)[2]; \
    (a)[2] = c; \
}

#define fftKernel4s(a0,a1,a2,a3,dir) \
{ \
    fftKernel2S((a0), (a2), dir); \
    fftKernel2S((a1), (a3), dir); \
    fftKernel2S((a0), (a1), dir); \
    (a3) = (float2)(dir)*(conjTransp((a3))); \
    fftKernel2S((a2), (a3), dir); \
    float2 c = (a1); \
    (a1) = (a2); \
    (a2) = c; \
}

#define bitreverse8(a) \
{ \
    float2 c; \
    c = (a)[1]; \
    (a)[1] = (a)[4]; \
    (a)[4] = c; \
    c = (a)[3]; \
    (a)[3] = (a)[6]; \
    (a)[6] = c; \
}

#define fftKernel8(a,dir) \
{ \
        const float2 w1  = (float2)(0x1.6a09e6p-1f,  dir*0x1.6a09e6p-1f);  \
        const float2 w3  = (float2)(-0x1.6a09e6p-1f, dir*0x1.6a09e6p-1f);  \
        float2 c; \
        fftKernel2S((a)[0], (a)[4], dir); \
        fftKernel2S((a)[1], (a)[5], dir); \
        fftKernel2S((a)[2], (a)[6], dir); \
        fftKernel2S((a)[3], (a)[7], dir); \
        (a)[5] = complexMul(w1, (a)[5]); \
        (a)[6] = (float2)(dir)*(conjTransp((a)[6])); \
        (a)[7] = complexMul(w3, (a)[7]); \
        fftKernel2S((a)[0], (a)[2], dir); \
        fftKernel2S((a)[1], (a)[3], dir); \
        fftKernel2S((a)[4], (a)[6], dir); \
        fftKernel2S((a)[5], (a)[7], dir); \
        (a)[3] = (float2)(dir)*(conjTransp((a)[3])); \
        (a)[7] = (float2)(dir)*(conjTransp((a)[7])); \
        fftKernel2S((a)[0], (a)[1], dir); \
        fftKernel2S((a)[2], (a)[3], dir); \
        fftKernel2S((a)[4], (a)[5], dir); \
        fftKernel2S((a)[6], (a)[7], dir); \
        bitreverse8((a)); \
}

#define bitreverse4x4(a) \
{ \
        float2 c; \
        c = (a)[1];  (a)[1]  = (a)[4];  (a)[4]  = c; \
        c = (a)[2];  (a)[2]  = (a)[8];  (a)[8]  = c; \
        c = (a)[3];  (a)[3]  = (a)[12]; (a)[12] = c; \
        c = (a)[6];  (a)[6]  = (a)[9];  (a)[9]  = c; \
        c = (a)[7];  (a)[7]  = (a)[13]; (a)[13] = c; \
        c = (a)[11]; (a)[11] = (a)[14]; (a)[14] = c; \
}

#define fftKernel16(a,dir) \
{ \
    const float w0 = 0x1.d906bcp-1f; \
    const float w1 = 0x1.87de2ap-2f; \
    const float w2 = 0x1.6a09e6p-1f; \
    fftKernel4s((a)[0], (a)[4], (a)[8],  (a)[12], dir); \
    fftKernel4s((a)[1], (a)[5], (a)[9],  (a)[13], dir); \
    fftKernel4s((a)[2], (a)[6], (a)[10], (a)[14], dir); \
    fftKernel4s((a)[3], (a)[7], (a)[11], (a)[15], dir); \
    (a)[5]  = complexMul((a)[5], (float2)(w0, dir*w1)); \
    (a)[6]  = complexMul((a)[6], (float2)(w2, dir*w2)); \
    (a)[7]  = complexMul((a)[7], (float2)(w1, dir*w0)); \
    (a)[9]  = complexMul((a)[9], (float2)(w2, dir*w2)); \
    (a)[10] = (float2)(dir)*(conjTransp((a)[10])); \
    (a)[11] = complexMul((a)[11], (float2)(-w2, dir*w2)); \
    (a)[13] = complexMul((a)[13], (float2)(w1, dir*w0)); \
    (a)[14] = complexMul((a)[14], (float2)(-w2, dir*w2)); \
    (a)[15] = complexMul((a)[15], (float2)(-w0, dir*-w1)); \
    fftKernel4((a), dir); \
    fftKernel4((a) + 4, dir); \
    fftKernel4((a) + 8, dir); \
    fftKernel4((a) + 12, dir); \
    bitreverse4x4((a)); \
}

#define bitreverse32(a) \
{ \
    float2 c1, c2; \
    c1 = (a)[2];   (a)[2] = (a)[1];   c2 = (a)[4];   (a)[4] = c1;   c1 = (a)[8];   (a)[8] = c2;    c2 = (a)[16];  (a)[16] = c1;   (a)[1] = c2; \
    c1 = (a)[6];   (a)[6] = (a)[3];   c2 = (a)[12];  (a)[12] = c1;  c1 = (a)[24];  (a)[24] = c2;   c2 = (a)[17];  (a)[17] = c1;   (a)[3] = c2; \
    c1 = (a)[10];  (a)[10] = (a)[5];  c2 = (a)[20];  (a)[20] = c1;  c1 = (a)[9];   (a)[9] = c2;    c2 = (a)[18];  (a)[18] = c1;   (a)[5] = c2; \
    c1 = (a)[14];  (a)[14] = (a)[7];  c2 = (a)[28];  (a)[28] = c1;  c1 = (a)[25];  (a)[25] = c2;   c2 = (a)[19];  (a)[19] = c1;   (a)[7] = c2; \
    c1 = (a)[22];  (a)[22] = (a)[11]; c2 = (a)[13];  (a)[13] = c1;  c1 = (a)[26];  (a)[26] = c2;   c2 = (a)[21];  (a)[21] = c1;   (a)[11] = c2; \
    c1 = (a)[30];  (a)[30] = (a)[15]; c2 = (a)[29];  (a)[29] = c1;  c1 = (a)[27];  (a)[27] = c2;   c2 = (a)[23];  (a)[23] = c1;   (a)[15] = c2; \
}

#define fftKernel32(a,dir) \
{ \
    fftKernel2S((a)[0],  (a)[16], dir); \
    fftKernel2S((a)[1],  (a)[17], dir); \
    fftKernel2S((a)[2],  (a)[18], dir); \
    fftKernel2S((a)[3],  (a)[19], dir); \
    fftKernel2S((a)[4],  (a)[20], dir); \
    fftKernel2S((a)[5],  (a)[21], dir); \
    fftKernel2S((a)[6],  (a)[22], dir); \
    fftKernel2S((a)[7],  (a)[23], dir); \
    fftKernel2S((a)[8],  (a)[24], dir); \
    fftKernel2S((a)[9],  (a)[25], dir); \
    fftKernel2S((a)[10], (a)[26], dir); \
    fftKernel2S((a)[11], (a)[27], dir); \
    fftKernel2S((a)[12], (a)[28], dir); \
    fftKernel2S((a)[13], (a)[29], dir); \
    fftKernel2S((a)[14], (a)[30], dir); \
    fftKernel2S((a)[15], (a)[31], dir); \
    (a)[17] = complexMul((a)[17], (float2)(0x1.f6297cp-1f, dir*0x1.8f8b84p-3f)); \
    (a)[18] = complexMul((a)[18], (float2)(0x1.d906bcp-1f, dir*0x1.87de2ap-2f)); \
    (a)[19] = complexMul((a)[19], (float2)(0x1.a9b662p-1f, dir*0x1.1c73b4p-1f)); \
    (a)[20] = complexMul((a)[20], (float2)(0x1.6a09e6p-1f, dir*0x1.6a09e6p-1f)); \
    (a)[21] = complexMul((a)[21], (float2)(0x1.1c73b4p-1f, dir*0x1.a9b662p-1f)); \
    (a)[22] = complexMul((a)[22], (float2)(0x1.87de2ap-2f, dir*0x1.d906bcp-1f)); \
    (a)[23] = complexMul((a)[23], (float2)(0x1.8f8b84p-3f, dir*0x1.f6297cp-1f)); \
    (a)[24] = complexMul((a)[24], (float2)(0x0p+0f, dir*0x1p+0f)); \
    (a)[25] = complexMul((a)[25], (float2)(-0x1.8f8b84p-3f, dir*0x1.f6297cp-1f)); \
    (a)[26] = complexMul((a)[26], (float2)(-0x1.87de2ap-2f, dir*0x1.d906bcp-1f)); \
    (a)[27] = complexMul((a)[27], (float2)(-0x1.1c73b4p-1f, dir*0x1.a9b662p-1f)); \
    (a)[28] = complexMul((a)[28], (float2)(-0x1.6a09e6p-1f, dir*0x1.6a09e6p-1f)); \
    (a)[29] = complexMul((a)[29], (float2)(-0x1.a9b662p-1f, dir*0x1.1c73b4p-1f)); \
    (a)[30] = complexMul((a)[30], (float2)(-0x1.d906bcp-1f, dir*0x1.87de2ap-2f)); \
    (a)[31] = complexMul((a)[31], (float2)(-0x1.f6297cp-1f, dir*0x1.8f8b84p-3f)); \
    fftKernel16((a), dir); \
    fftKernel16((a) + 16, dir); \
    bitreverse32((a)); \
}

__kernel void \
clFFT_1DTwistInterleaved(__global float2 *in, unsigned int startRow, unsigned int numCols, unsigned int N, unsigned int numRowsToProcess, int dir) \
{ \
   float2 a, w; \
   float ang; \
   unsigned int j; \
        unsigned int i = get_global_id(0); \
        unsigned int startIndex = i; \
         \
        if(i < numCols) \
        { \
            for(j = 0; j < numRowsToProcess; j++) \
            { \
                a = in[startIndex]; \
                ang = 2.0f * M_PI * dir * i * (startRow + j) / N; \
                w = (float2)(native_cos(ang), native_sin(ang)); \
                a = complexMul(a, w); \
                in[startIndex] = a; \
                startIndex += numCols; \
            } \
        }        \
} \

/**
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060, 
 * for 1D input of length 1024, with modified radices (8 - 8 - 4 - 4) for ATI compatability
 */ 
__kernel void clFFT_custom(__global float *speex_in, unsigned int speex_in_offset, __global float2 *in, __global float2 *out, __global float *speex_out, unsigned int speex_out_offset, int dir, int S)
{
	__local float sMem[1088];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[8];
	int lId = get_local_id(0);
	int groupId = get_group_id(0);
	ii = lId;
	jj = 0;

	offset =  mad24(groupId, 1024, ii);
	//offset = ii;
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	int workItemOffset = lId * 8 + groupId * 1024;
	
	// Do preparation step to convert from Speex FFT format 
	// There is one thread for each four positions in the FFT length

	// clFFT_Forward
	if(dir == -1)
	{	
		for(int t=0;t<8;t++)
		{
			in[workItemOffset + t].x = speex_in[workItemOffset + t + speex_in_offset] * (1.0f / 1024.0f);
			in[workItemOffset + t].y = 0.0f;
		}
		
		barrier(CLK_GLOBAL_MEM_FENCE);
	        barrier(CLK_LOCAL_MEM_FENCE);
	}
	// clFFT_Inverse
	else if(dir == 1)
	{
		int groupOffset = groupId * 1024;
		if(lId == 0)
		{
			in[groupOffset + 0].x = speex_in[groupOffset + speex_in_offset + 0];
			in[groupOffset + 0].y = 0.0f;
			for(int t=1;t<4;t++)
			{
				in[groupOffset + t].x = speex_in[groupOffset + speex_in_offset + (2*t - 1)];
				in[groupOffset + t].y = speex_in[groupOffset + speex_in_offset + (2*t - 1) + 1];
			}
			in[groupOffset + 512].x = speex_in[groupOffset + speex_in_offset + 1024 - 1];
			in[groupOffset + 512].y = 0.0f;
		}
		else
		{
			for(int t=0;t<4;t++)
			{
				in[groupOffset + 4*lId + t].x = speex_in[groupOffset + speex_in_offset + lId*8 - 1 + 2*t];
				in[groupOffset + 4*lId + t].y = speex_in[groupOffset + speex_in_offset + lId*8 + 2*t];
				//in[groupOffset + 8*lId + t].x = 9000.0f;
				//in[groupOffset + 8*lId + t].y = 9000.0f;
			}
		}

		barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

		// Generate negative frequencies for symmetric input 
	        // Source: http://www.dspguide.com/ch12/1.htm
	        // For the remaining n/2 threads that needs to mirror 
		for(int t=0;t<4;t++)
		{
			in[groupOffset + 512 + lId*4 + t].x = in[groupOffset + 1024 - (512 + lId*4 + t)].x;
			in[groupOffset + 512 + lId*4 + t].y = -in[groupOffset + 1024 - (512 + lId*4 + t)].y;
		}
	}

	barrier(CLK_LOCAL_MEM_FENCE);
	
	in += offset;
	out += offset;
	a[0] = in[0];
	a[1] = in[128];
	a[2] = in[256];
	a[3] = in[384];
	a[4] = in[512];
	a[5] = in[640];
	a[6] = in[768];
	a[7] = in[896];
	fftKernel8(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii & 7;
	i = ii >> 3;
	lMemLoad = sMem + mad24(j, 130, i);
	lMemStore[0] = a[0].x;
	lMemStore[130] = a[1].x;
	lMemStore[260] = a[2].x;
	lMemStore[390] = a[3].x;
	lMemStore[520] = a[4].x;
	lMemStore[650] = a[5].x;
	lMemStore[780] = a[6].x;
	lMemStore[910] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	a[4].x = lMemLoad[64];
	a[5].x = lMemLoad[80];
	a[6].x = lMemLoad[96];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[130] = a[1].y;
	lMemStore[260] = a[2].y;
	lMemStore[390] = a[3].y;
	lMemStore[520] = a[4].y;
	lMemStore[650] = a[5].y;
	lMemStore[780] = a[6].y;
	lMemStore[910] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	a[4].y = lMemLoad[64];
	a[5].y = lMemLoad[80];
	a[6].y = lMemLoad[96];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel8(a+0, dir);
	angf = (float) (ii >> 3);
	ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = (ii & 63) >> 3;
	i = mad24(ii >> 6, 8, ii & 7);
	lMemLoad = sMem + mad24(j, 136, i);
	lMemStore[0] = a[0].x;
	lMemStore[136] = a[1].x;
	lMemStore[272] = a[2].x;
	lMemStore[408] = a[3].x;
	lMemStore[544] = a[4].x;
	lMemStore[680] = a[5].x;
	lMemStore[816] = a[6].x;
	lMemStore[952] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[32];
	a[2].x = lMemLoad[64];
	a[3].x = lMemLoad[96];
	a[4].x = lMemLoad[16];
	a[5].x = lMemLoad[48];
	a[6].x = lMemLoad[80];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[136] = a[1].y;
	lMemStore[272] = a[2].y;
	lMemStore[408] = a[3].y;
	lMemStore[544] = a[4].y;
	lMemStore[680] = a[5].y;
	lMemStore[816] = a[6].y;
	lMemStore[952] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[32];
	a[2].y = lMemLoad[64];
	a[3].y = lMemLoad[96];
	a[4].y = lMemLoad[16];
	a[5].y = lMemLoad[48];
	a[6].y = lMemLoad[80];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	angf = (float) (ii >> 6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	angf = (float) ((128 + ii) >>6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii >> 6;
	i = ii & 63;
	lMemLoad = sMem + mad24(j, 256, i);
	lMemStore[0] = a[0].x;
	lMemStore[256] = a[1].x;
	lMemStore[512] = a[2].x;
	lMemStore[768] = a[3].x;
	lMemStore[128] = a[4].x;
	lMemStore[384] = a[5].x;
	lMemStore[640] = a[6].x;
	lMemStore[896] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[64];
	a[2].x = lMemLoad[128];
	a[3].x = lMemLoad[192];
	a[4].x = lMemLoad[512];
	a[5].x = lMemLoad[576];
	a[6].x = lMemLoad[640];
	a[7].x = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[256] = a[1].y;
	lMemStore[512] = a[2].y;
	lMemStore[768] = a[3].y;
	lMemStore[128] = a[4].y;
	lMemStore[384] = a[5].y;
	lMemStore[640] = a[6].y;
	lMemStore[896] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[64];
	a[2].y = lMemLoad[128];
	a[3].y = lMemLoad[192];
	a[4].y = lMemLoad[512];
	a[5].y = lMemLoad[576];
	a[6].y = lMemLoad[640];
	a[7].y = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	out[0] = a[0];
	out[128] = a[4];
	out[256] = a[1];
	out[384] = a[5];
	out[512] = a[2];
	out[640] = a[6];
	out[768] = a[3];
	out[896] = a[7];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length

	// clFFT_Forward
	if(dir == -1)
	{		
		int workItemOffset2 = lId * 4 + groupId*1024;

		// The first thread for each FFT length
		if(lId == 0)
		{
			speex_out[speex_out_offset + groupId*1024 + 0] = out_orig[workItemOffset2 + 0].x;
			speex_out[speex_out_offset + groupId*1024 + 1] = out_orig[workItemOffset2 + 1].x;
			for(int t=1;t<4;t++)
			{
				speex_out[speex_out_offset + groupId*1024 + 2*t] = out_orig[workItemOffset2 + t].y;
				speex_out[speex_out_offset + groupId*1024 + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
			}
		}
		else
		{
			for(int t=0;t<4;t++)
			{
				speex_out[speex_out_offset + workItemOffset + 2*t] = out_orig[workItemOffset2 + t].y;
				speex_out[speex_out_offset + workItemOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
			}
		}
	}
	// clFFT_Inverse
	else if(dir == 1)
	{
		for(int t=0;t<8;t++)
		{
			speex_out[speex_out_offset + workItemOffset + t] = out_orig[workItemOffset + t].x;
		}
	}
	
	// End of finalization step
}

/**
 * Specialized kernel for updating the filter weights in batches
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 1024.
 */
__kernel void clFFT_update_weights(__global float *W, __global float2 *in, __global float2 *out, __global float *wtmp, int S)
{
	
	__local float sMem[1088];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	int chan = get_global_id(1);
	int speak = get_global_id(2);
	int num_speakers = get_global_size(2);
	int num_channels = get_global_size(1);
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[8];
	int lId = get_local_id( 0 );
	int groupId = get_group_id( 0 );
	const float m = 1.0f / 1024.0f;
	int groupOffset = groupId * 1024 * num_speakers;	// Group offset for float
	int groupOffset2 = groupId * 512 * num_speakers;	// Group offset for float2
	unsigned int cs_offset = chan*1024*num_speakers*S + speak*1024; // Channel/speaker offset
	unsigned int W_offset = cs_offset;
	ii = lId;
	jj = 0;

	// Do IFFT first
	int dir = 1;
	
	//offset =  mad24(groupId, 1024, ii);
	offset = cs_offset + groupId * 1024 * num_speakers + ii;
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	
	int workItemOffset = lId * 8 + groupOffset + cs_offset;
	
	barrier(CLK_LOCAL_MEM_FENCE);

	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	if(lId == 0)
	{		
		in[groupOffset + cs_offset + 0].x = W[groupOffset + cs_offset + 0];
		in[groupOffset + cs_offset + 0].y = 0.0f;
		for(int t=1;t<4;t++)
		{
			in[groupOffset + cs_offset + t].x = W[groupOffset + cs_offset + (2*t - 1)];
			in[groupOffset + cs_offset + t].y = W[groupOffset + cs_offset + (2*t - 1) + 1];
		}
		in[groupOffset + cs_offset + 512].x = W[groupOffset + cs_offset + 1024 - 1];
		in[groupOffset + cs_offset + 512].y = 0.0f;
	}
	else
	{
		for(int t=0;t<4;t++)
		{
			in[groupOffset + cs_offset + 4*lId + t].x = W[groupOffset + cs_offset + lId*8 - 1 + 2*t];
			in[groupOffset + cs_offset + 4*lId + t].y = W[groupOffset + cs_offset + lId*8 + 2*t];
		}
	}

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	// Generate negative frequencies for symmetric input 
	// Source: http://www.dspguide.com/ch12/1.htm
	// For the remaining n/2 threads that needs to mirror
	for(int t=0;t<4;t++)
	{
		in[groupOffset + cs_offset + 512 + lId*4 + t].x = in[groupOffset + cs_offset + 1024 - (512 + lId*4 + t)].x;
		in[groupOffset + cs_offset + 512 + lId*4 + t].y = -in[groupOffset + cs_offset + 1024 - (512 + lId*4 + t)].y;
	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	in += offset;
	out += offset;
	a[0] = in[0];
	a[1] = in[128];
	a[2] = in[256];
	a[3] = in[384];
	a[4] = in[512];
	a[5] = in[640];
	a[6] = in[768];
	a[7] = in[896];
	fftKernel8(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii & 7;
	i = ii >> 3;
	lMemLoad = sMem + mad24(j, 130, i);
	lMemStore[0] = a[0].x;
	lMemStore[130] = a[1].x;
	lMemStore[260] = a[2].x;
	lMemStore[390] = a[3].x;
	lMemStore[520] = a[4].x;
	lMemStore[650] = a[5].x;
	lMemStore[780] = a[6].x;
	lMemStore[910] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	a[4].x = lMemLoad[64];
	a[5].x = lMemLoad[80];
	a[6].x = lMemLoad[96];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[130] = a[1].y;
	lMemStore[260] = a[2].y;
	lMemStore[390] = a[3].y;
	lMemStore[520] = a[4].y;
	lMemStore[650] = a[5].y;
	lMemStore[780] = a[6].y;
	lMemStore[910] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	a[4].y = lMemLoad[64];
	a[5].y = lMemLoad[80];
	a[6].y = lMemLoad[96];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel8(a+0, dir);
	angf = (float) (ii >> 3);
	ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = (ii & 63) >> 3;
	i = mad24(ii >> 6, 8, ii & 7);
	lMemLoad = sMem + mad24(j, 136, i);
	lMemStore[0] = a[0].x;
	lMemStore[136] = a[1].x;
	lMemStore[272] = a[2].x;
	lMemStore[408] = a[3].x;
	lMemStore[544] = a[4].x;
	lMemStore[680] = a[5].x;
	lMemStore[816] = a[6].x;
	lMemStore[952] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[32];
	a[2].x = lMemLoad[64];
	a[3].x = lMemLoad[96];
	a[4].x = lMemLoad[16];
	a[5].x = lMemLoad[48];
	a[6].x = lMemLoad[80];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[136] = a[1].y;
	lMemStore[272] = a[2].y;
	lMemStore[408] = a[3].y;
	lMemStore[544] = a[4].y;
	lMemStore[680] = a[5].y;
	lMemStore[816] = a[6].y;
	lMemStore[952] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[32];
	a[2].y = lMemLoad[64];
	a[3].y = lMemLoad[96];
	a[4].y = lMemLoad[16];
	a[5].y = lMemLoad[48];
	a[6].y = lMemLoad[80];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	angf = (float) (ii >> 6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	angf = (float) ((128 + ii) >>6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii >> 6;
	i = ii & 63;
	lMemLoad = sMem + mad24(j, 256, i);
	lMemStore[0] = a[0].x;
	lMemStore[256] = a[1].x;
	lMemStore[512] = a[2].x;
	lMemStore[768] = a[3].x;
	lMemStore[128] = a[4].x;
	lMemStore[384] = a[5].x;
	lMemStore[640] = a[6].x;
	lMemStore[896] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[64];
	a[2].x = lMemLoad[128];
	a[3].x = lMemLoad[192];
	a[4].x = lMemLoad[512];
	a[5].x = lMemLoad[576];
	a[6].x = lMemLoad[640];
	a[7].x = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[256] = a[1].y;
	lMemStore[512] = a[2].y;
	lMemStore[768] = a[3].y;
	lMemStore[128] = a[4].y;
	lMemStore[384] = a[5].y;
	lMemStore[640] = a[6].y;
	lMemStore[896] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[64];
	a[2].y = lMemLoad[128];
	a[3].y = lMemLoad[192];
	a[4].y = lMemLoad[512];
	a[5].y = lMemLoad[576];
	a[6].y = lMemLoad[640];
	a[7].y = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	out[0] = a[0];
	out[128] = a[4];
	out[256] = a[1];
	out[384] = a[5];
	out[512] = a[2];
	out[640] = a[6];
	out[768] = a[3];
	out[896] = a[7];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length
	for(int t=0;t<8;t++)
	{
		wtmp[workItemOffset + t] = out_orig[workItemOffset + t].x;
		//wtmp[workItemOffset + t] = 1.0f;
	}

	// End of finalization step
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	//barrier(CLK_LOCAL_MEM_FENCE);
	
	//workItemOffset = lId * 8 + groupOffset + cs_offset;
	
	// Set to zero 
	wtmp[groupOffset + cs_offset + 512 + 4*lId] = 0.0f;
	wtmp[groupOffset + cs_offset + 512 + 4*lId + 1] = 0.0f;
	wtmp[groupOffset + cs_offset + 512 + 4*lId + 2] = 0.0f;
	wtmp[groupOffset + cs_offset + 512 + 4*lId + 3] = 0.0f;
	
	in = in_orig;
	out = out_orig;
	
	// Do forward FFT
	dir = -1;
	
	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	// clFFT_Forward
	for(int t=0;t<8;t++)
	{
		in_orig[workItemOffset + t].x = wtmp[workItemOffset + t] * m;
		//in_orig[workItemOffset + t].x = 1.0f;
		in_orig[workItemOffset + t].y = 0.0f;
	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	
	// End of preparation step
	
	offset = cs_offset + groupId * 1024 * num_speakers + ii;
	in += offset;
	out += offset;
	a[0] = in[0];
	a[1] = in[128];
	a[2] = in[256];
	a[3] = in[384];
	a[4] = in[512];
	a[5] = in[640];
	a[6] = in[768];
	a[7] = in[896];
	fftKernel8(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii & 7;
	i = ii >> 3;
	lMemLoad = sMem + mad24(j, 130, i);
	lMemStore[0] = a[0].x;
	lMemStore[130] = a[1].x;
	lMemStore[260] = a[2].x;
	lMemStore[390] = a[3].x;
	lMemStore[520] = a[4].x;
	lMemStore[650] = a[5].x;
	lMemStore[780] = a[6].x;
	lMemStore[910] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	a[4].x = lMemLoad[64];
	a[5].x = lMemLoad[80];
	a[6].x = lMemLoad[96];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[130] = a[1].y;
	lMemStore[260] = a[2].y;
	lMemStore[390] = a[3].y;
	lMemStore[520] = a[4].y;
	lMemStore[650] = a[5].y;
	lMemStore[780] = a[6].y;
	lMemStore[910] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	a[4].y = lMemLoad[64];
	a[5].y = lMemLoad[80];
	a[6].y = lMemLoad[96];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel8(a+0, dir);
	angf = (float) (ii >> 3);
	ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = (ii & 63) >> 3;
	i = mad24(ii >> 6, 8, ii & 7);
	lMemLoad = sMem + mad24(j, 136, i);
	lMemStore[0] = a[0].x;
	lMemStore[136] = a[1].x;
	lMemStore[272] = a[2].x;
	lMemStore[408] = a[3].x;
	lMemStore[544] = a[4].x;
	lMemStore[680] = a[5].x;
	lMemStore[816] = a[6].x;
	lMemStore[952] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[32];
	a[2].x = lMemLoad[64];
	a[3].x = lMemLoad[96];
	a[4].x = lMemLoad[16];
	a[5].x = lMemLoad[48];
	a[6].x = lMemLoad[80];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[136] = a[1].y;
	lMemStore[272] = a[2].y;
	lMemStore[408] = a[3].y;
	lMemStore[544] = a[4].y;
	lMemStore[680] = a[5].y;
	lMemStore[816] = a[6].y;
	lMemStore[952] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[32];
	a[2].y = lMemLoad[64];
	a[3].y = lMemLoad[96];
	a[4].y = lMemLoad[16];
	a[5].y = lMemLoad[48];
	a[6].y = lMemLoad[80];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	angf = (float) (ii >> 6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	angf = (float) ((128 + ii) >>6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii >> 6;
	i = ii & 63;
	lMemLoad = sMem + mad24(j, 256, i);
	lMemStore[0] = a[0].x;
	lMemStore[256] = a[1].x;
	lMemStore[512] = a[2].x;
	lMemStore[768] = a[3].x;
	lMemStore[128] = a[4].x;
	lMemStore[384] = a[5].x;
	lMemStore[640] = a[6].x;
	lMemStore[896] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[64];
	a[2].x = lMemLoad[128];
	a[3].x = lMemLoad[192];
	a[4].x = lMemLoad[512];
	a[5].x = lMemLoad[576];
	a[6].x = lMemLoad[640];
	a[7].x = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[256] = a[1].y;
	lMemStore[512] = a[2].y;
	lMemStore[768] = a[3].y;
	lMemStore[128] = a[4].y;
	lMemStore[384] = a[5].y;
	lMemStore[640] = a[6].y;
	lMemStore[896] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[64];
	a[2].y = lMemLoad[128];
	a[3].y = lMemLoad[192];
	a[4].y = lMemLoad[512];
	a[5].y = lMemLoad[576];
	a[6].y = lMemLoad[640];
	a[7].y = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	out[0] = a[0];
	out[128] = a[4];
	out[256] = a[1];
	out[384] = a[5];
	out[512] = a[2];
	out[640] = a[6];
	out[768] = a[3];
	out[896] = a[7];
	
	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
	
	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length
	// Not adapted to multi-channel filter, only mono
	// groupOffset = groupId * 256 * num_speakers;
	int workItemOffset2 = lId * 4 + groupOffset + cs_offset;
	//workItemOffset = lId * 2 + groupOffset;
	//unsigned int W_offset = speak*1024 + chan*1024*num_speakers*S;
	
	// The first thread for each FFT length
	if(lId == 0)
	{		
		W[cs_offset + groupOffset + 0] = out_orig[workItemOffset2 + 0].x;
		W[cs_offset + groupOffset + 1] = out_orig[workItemOffset2 + 1].x;
		for(int t=1;t<4;t++)
		{
			W[cs_offset + groupOffset + 2*t] = out_orig[workItemOffset2 + t].y;
			W[cs_offset + groupOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
		}
	}
	// note: cs_offset is included in workItemOffset
	else if(groupId < S)
	{
		for(int t=0;t<4;t++)
		{
			W[workItemOffset + 2*t] = out_orig[workItemOffset2 + t].y;
			W[workItemOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;
			//W[workItemOffset + 2*t] = 9000.0f;
			//W[workItemOffset + 2*t + 1] = 8000.0f;
			//W[workItemOffset + 2*t] = wtmp[groupOffset + cs_offset + 8*lId + t];
			//W[workItemOffset + 2*t + 1] = wtmp[groupOffset + cs_offset + 8*lId + t];
		}
	}
	
	// End of finalization step 
	
}

/**
 * Specialized kernel for updating the filter weights in batches
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 1024.
 * Split in two for ATI, as it crashes on a single long kernel
 */
__kernel void clFFT_update_weights_ati_pt1(__global float *W, __global float2 *in, __global float2 *out, __global float *wtmp, int S)
{
	
	__local float sMem[1088];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	int chan = get_global_id(1);
	int speak = get_global_id(2);
	int num_speakers = get_global_size(2);
	int num_channels = get_global_size(1);
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[8];
	int lId = get_local_id( 0 );
	int groupId = get_group_id( 0 );
	const float m = 1.0f / 1024.0f;
	int groupOffset = groupId * 1024 * num_speakers;	// Group offset for float
	int groupOffset2 = groupId * 512 * num_speakers;	// Group offset for float2
	unsigned int cs_offset = chan*1024*num_speakers*S + speak*1024; // Channel/speaker offset
	unsigned int W_offset = cs_offset;
	ii = lId;
	jj = 0;

	// Do IFFT first
	int dir = 1;
	
	//offset =  mad24(groupId, 1024, ii);
	offset = cs_offset + groupId * 1024 * num_speakers + ii;
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	
	int workItemOffset = lId * 8 + groupOffset + cs_offset;
	
	barrier(CLK_LOCAL_MEM_FENCE);

	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	if(lId == 0)
	{		
		in[groupOffset + cs_offset + 0].x = W[groupOffset + cs_offset + 0];
		in[groupOffset + cs_offset + 0].y = 0.0f;
		for(int t=1;t<4;t++)
		{
			in[groupOffset + cs_offset + t].x = W[groupOffset + cs_offset + (2*t - 1)];
			in[groupOffset + cs_offset + t].y = W[groupOffset + cs_offset + (2*t - 1) + 1];
		}
		in[groupOffset + cs_offset + 512].x = W[groupOffset + cs_offset + 1024 - 1];
		in[groupOffset + cs_offset + 512].y = 0.0f;
	}
	else
	{
		for(int t=0;t<4;t++)
		{
			in[groupOffset + cs_offset + 4*lId + t].x = W[groupOffset + cs_offset + lId*8 - 1 + 2*t];
			in[groupOffset + cs_offset + 4*lId + t].y = W[groupOffset + cs_offset + lId*8 + 2*t];
		}
	}

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	// Generate negative frequencies for symmetric input 
	// Source: http://www.dspguide.com/ch12/1.htm
	// For the remaining n/2 threads that needs to mirror
	for(int t=0;t<4;t++)
	{
		in[groupOffset + cs_offset + 512 + lId*4 + t].x = in[groupOffset + cs_offset + 1024 - (512 + lId*4 + t)].x;
		in[groupOffset + cs_offset + 512 + lId*4 + t].y = -in[groupOffset + cs_offset + 1024 - (512 + lId*4 + t)].y;
	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	in += offset;
	out += offset;
	a[0] = in[0];
	a[1] = in[128];
	a[2] = in[256];
	a[3] = in[384];
	a[4] = in[512];
	a[5] = in[640];
	a[6] = in[768];
	a[7] = in[896];
	fftKernel8(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii & 7;
	i = ii >> 3;
	lMemLoad = sMem + mad24(j, 130, i);
	lMemStore[0] = a[0].x;
	lMemStore[130] = a[1].x;
	lMemStore[260] = a[2].x;
	lMemStore[390] = a[3].x;
	lMemStore[520] = a[4].x;
	lMemStore[650] = a[5].x;
	lMemStore[780] = a[6].x;
	lMemStore[910] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	a[4].x = lMemLoad[64];
	a[5].x = lMemLoad[80];
	a[6].x = lMemLoad[96];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[130] = a[1].y;
	lMemStore[260] = a[2].y;
	lMemStore[390] = a[3].y;
	lMemStore[520] = a[4].y;
	lMemStore[650] = a[5].y;
	lMemStore[780] = a[6].y;
	lMemStore[910] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	a[4].y = lMemLoad[64];
	a[5].y = lMemLoad[80];
	a[6].y = lMemLoad[96];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel8(a+0, dir);
	angf = (float) (ii >> 3);
	ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = (ii & 63) >> 3;
	i = mad24(ii >> 6, 8, ii & 7);
	lMemLoad = sMem + mad24(j, 136, i);
	lMemStore[0] = a[0].x;
	lMemStore[136] = a[1].x;
	lMemStore[272] = a[2].x;
	lMemStore[408] = a[3].x;
	lMemStore[544] = a[4].x;
	lMemStore[680] = a[5].x;
	lMemStore[816] = a[6].x;
	lMemStore[952] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[32];
	a[2].x = lMemLoad[64];
	a[3].x = lMemLoad[96];
	a[4].x = lMemLoad[16];
	a[5].x = lMemLoad[48];
	a[6].x = lMemLoad[80];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[136] = a[1].y;
	lMemStore[272] = a[2].y;
	lMemStore[408] = a[3].y;
	lMemStore[544] = a[4].y;
	lMemStore[680] = a[5].y;
	lMemStore[816] = a[6].y;
	lMemStore[952] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[32];
	a[2].y = lMemLoad[64];
	a[3].y = lMemLoad[96];
	a[4].y = lMemLoad[16];
	a[5].y = lMemLoad[48];
	a[6].y = lMemLoad[80];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	angf = (float) (ii >> 6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	angf = (float) ((128 + ii) >>6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii >> 6;
	i = ii & 63;
	lMemLoad = sMem + mad24(j, 256, i);
	lMemStore[0] = a[0].x;
	lMemStore[256] = a[1].x;
	lMemStore[512] = a[2].x;
	lMemStore[768] = a[3].x;
	lMemStore[128] = a[4].x;
	lMemStore[384] = a[5].x;
	lMemStore[640] = a[6].x;
	lMemStore[896] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[64];
	a[2].x = lMemLoad[128];
	a[3].x = lMemLoad[192];
	a[4].x = lMemLoad[512];
	a[5].x = lMemLoad[576];
	a[6].x = lMemLoad[640];
	a[7].x = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[256] = a[1].y;
	lMemStore[512] = a[2].y;
	lMemStore[768] = a[3].y;
	lMemStore[128] = a[4].y;
	lMemStore[384] = a[5].y;
	lMemStore[640] = a[6].y;
	lMemStore[896] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[64];
	a[2].y = lMemLoad[128];
	a[3].y = lMemLoad[192];
	a[4].y = lMemLoad[512];
	a[5].y = lMemLoad[576];
	a[6].y = lMemLoad[640];
	a[7].y = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	out[0] = a[0];
	out[128] = a[4];
	out[256] = a[1];
	out[384] = a[5];
	out[512] = a[2];
	out[640] = a[6];
	out[768] = a[3];
	out[896] = a[7];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length
	for(int t=0;t<8;t++)
	{
		wtmp[workItemOffset + t] = out_orig[workItemOffset + t].x;
		//wtmp[workItemOffset + t] = 1.0f;
	}

	// End of finalization step
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	//barrier(CLK_LOCAL_MEM_FENCE);
	
	// Set to zero 
	wtmp[groupOffset + cs_offset + 512 + 4*lId] = 0.0f;
	wtmp[groupOffset + cs_offset + 512 + 4*lId + 1] = 0.0f;
	wtmp[groupOffset + cs_offset + 512 + 4*lId + 2] = 0.0f;
	wtmp[groupOffset + cs_offset + 512 + 4*lId + 3] = 0.0f;
}

/**
 * Specialized kernel for updating the filter weights in batches
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 1024.
 * Split in two for ATI, as it crashes on a single long kernel
 */
__kernel void clFFT_update_weights_ati_pt2(__global float *W, __global float2 *in, __global float2 *out, __global float *wtmp, int S)
{
	
	__local float sMem[1088];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	int chan = get_global_id(1);
	int speak = get_global_id(2);
	int num_speakers = get_global_size(2);
	int num_channels = get_global_size(1);
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[8];
	int lId = get_local_id( 0 );
	int groupId = get_group_id( 0 );
	const float m = 1.0f / 1024.0f;
	int groupOffset = groupId * 1024 * num_speakers;	// Group offset for float
	int groupOffset2 = groupId * 512 * num_speakers;	// Group offset for float2
	unsigned int cs_offset = chan*1024*num_speakers*S + speak*1024; // Channel/speaker offset
	unsigned int W_offset = cs_offset;
	ii = lId;
	jj = 0;

	// Do IFFT first
	int dir = 1;
	
	//offset =  mad24(groupId, 1024, ii);
	offset = cs_offset + groupId * 1024 * num_speakers + ii;
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	
	int workItemOffset = lId * 8 + groupOffset + cs_offset;
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	// Do forward FFT
	dir = -1;
	
	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	// clFFT_Forward
	for(int t=0;t<8;t++)
	{
		in_orig[workItemOffset + t].x = wtmp[workItemOffset + t] * m;
		//in_orig[workItemOffset + t].x = 1.0f;
		in_orig[workItemOffset + t].y = 0.0f;
	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	
	// End of preparation step
	
	offset = cs_offset + groupId * 1024 * num_speakers + ii;
	in += offset;
	out += offset;
	a[0] = in[0];
	a[1] = in[128];
	a[2] = in[256];
	a[3] = in[384];
	a[4] = in[512];
	a[5] = in[640];
	a[6] = in[768];
	a[7] = in[896];
	fftKernel8(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii & 7;
	i = ii >> 3;
	lMemLoad = sMem + mad24(j, 130, i);
	lMemStore[0] = a[0].x;
	lMemStore[130] = a[1].x;
	lMemStore[260] = a[2].x;
	lMemStore[390] = a[3].x;
	lMemStore[520] = a[4].x;
	lMemStore[650] = a[5].x;
	lMemStore[780] = a[6].x;
	lMemStore[910] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	a[4].x = lMemLoad[64];
	a[5].x = lMemLoad[80];
	a[6].x = lMemLoad[96];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[130] = a[1].y;
	lMemStore[260] = a[2].y;
	lMemStore[390] = a[3].y;
	lMemStore[520] = a[4].y;
	lMemStore[650] = a[5].y;
	lMemStore[780] = a[6].y;
	lMemStore[910] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	a[4].y = lMemLoad[64];
	a[5].y = lMemLoad[80];
	a[6].y = lMemLoad[96];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel8(a+0, dir);
	angf = (float) (ii >> 3);
	ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = (ii & 63) >> 3;
	i = mad24(ii >> 6, 8, ii & 7);
	lMemLoad = sMem + mad24(j, 136, i);
	lMemStore[0] = a[0].x;
	lMemStore[136] = a[1].x;
	lMemStore[272] = a[2].x;
	lMemStore[408] = a[3].x;
	lMemStore[544] = a[4].x;
	lMemStore[680] = a[5].x;
	lMemStore[816] = a[6].x;
	lMemStore[952] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[32];
	a[2].x = lMemLoad[64];
	a[3].x = lMemLoad[96];
	a[4].x = lMemLoad[16];
	a[5].x = lMemLoad[48];
	a[6].x = lMemLoad[80];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[136] = a[1].y;
	lMemStore[272] = a[2].y;
	lMemStore[408] = a[3].y;
	lMemStore[544] = a[4].y;
	lMemStore[680] = a[5].y;
	lMemStore[816] = a[6].y;
	lMemStore[952] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[32];
	a[2].y = lMemLoad[64];
	a[3].y = lMemLoad[96];
	a[4].y = lMemLoad[16];
	a[5].y = lMemLoad[48];
	a[6].y = lMemLoad[80];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	angf = (float) (ii >> 6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	angf = (float) ((128 + ii) >>6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii >> 6;
	i = ii & 63;
	lMemLoad = sMem + mad24(j, 256, i);
	lMemStore[0] = a[0].x;
	lMemStore[256] = a[1].x;
	lMemStore[512] = a[2].x;
	lMemStore[768] = a[3].x;
	lMemStore[128] = a[4].x;
	lMemStore[384] = a[5].x;
	lMemStore[640] = a[6].x;
	lMemStore[896] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[64];
	a[2].x = lMemLoad[128];
	a[3].x = lMemLoad[192];
	a[4].x = lMemLoad[512];
	a[5].x = lMemLoad[576];
	a[6].x = lMemLoad[640];
	a[7].x = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[256] = a[1].y;
	lMemStore[512] = a[2].y;
	lMemStore[768] = a[3].y;
	lMemStore[128] = a[4].y;
	lMemStore[384] = a[5].y;
	lMemStore[640] = a[6].y;
	lMemStore[896] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[64];
	a[2].y = lMemLoad[128];
	a[3].y = lMemLoad[192];
	a[4].y = lMemLoad[512];
	a[5].y = lMemLoad[576];
	a[6].y = lMemLoad[640];
	a[7].y = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	out[0] = a[0];
	out[128] = a[4];
	out[256] = a[1];
	out[384] = a[5];
	out[512] = a[2];
	out[640] = a[6];
	out[768] = a[3];
	out[896] = a[7];
	
	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
	
	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length
	// Not adapted to multi-channel filter, only mono
	// groupOffset = groupId * 256 * num_speakers;
	int workItemOffset2 = lId * 4 + groupOffset + cs_offset;
	//workItemOffset = lId * 2 + groupOffset;
	//unsigned int W_offset = speak*1024 + chan*1024*num_speakers*S;
	
	// The first thread for each FFT length
	if(lId == 0)
	{		
		W[cs_offset + groupOffset + 0] = out_orig[workItemOffset2 + 0].x;
		W[cs_offset + groupOffset + 1] = out_orig[workItemOffset2 + 1].x;
		for(int t=1;t<4;t++)
		{
			W[cs_offset + groupOffset + 2*t] = out_orig[workItemOffset2 + t].y;
			W[cs_offset + groupOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
		}
	}
	// note: cs_offset is included in workItemOffset
	else if(groupId < S)
	{
		for(int t=0;t<4;t++)
		{
			W[workItemOffset + 2*t] = out_orig[workItemOffset2 + t].y;
			W[workItemOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;
			//W[workItemOffset + 2*t] = 9000.0f;
			//W[workItemOffset + 2*t + 1] = 8000.0f;
			//W[workItemOffset + 2*t] = wtmp[groupOffset + cs_offset + 8*lId + t];
			//W[workItemOffset + 2*t + 1] = wtmp[groupOffset + cs_offset + 8*lId + t];
		}
	}
	
	// End of finalization step 
	
}

/**
 * Specialized kernel for converting the error to the frequency domain in one operation 
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 1024.
 */
__kernel void clFFT_convert_error_to_freq_domain(__global float *e, __global float *E, __global float *y, __global float *Y, __global float2 *in, __global float2 *in2, __global float2 *out, __global float2 *out2, int S)
{
	
	__local float sMem[1088];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[8];
	int lId = get_local_id(0);
	int taskId = get_global_id(1);
	int groupId = get_group_id(0);
	const float m = 1.0f / 1024.0f;
	int groupOffset = groupId * 1024;	// Group offset for float
	int groupOffset2 = groupId * 512;	// Group offset for float2
	ii = lId;
	jj = 0;

	offset =  mad24(groupId, 1024, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	__global float2 *out2_orig = out2;	// Keep this pointer for finalize code
	__global float2 *in2_orig = in2;
	
	int workItemOffset = lId * 8 + groupOffset;

	barrier(CLK_LOCAL_MEM_FENCE);

	// Do FFT
	// Equiv: speex_echo_opencl_fft_execute2(ocl_e, chan*st->window_size, ocl_E, chan*st->window_size, 1, commandQueue);
	if(taskId == 0)
	{
		// Do FFT first
		int dir = -1;

		// Do preparation step to convert from Speex FFT format
		// There is one thread for each four positions in the FFT length

		// clFFT_Forward
		for(int t=0;t<8;t++)
		{
			in_orig[workItemOffset + t].x = e[workItemOffset + t] * m;
			in_orig[workItemOffset + t].y = 0.0f;
		}
		
		// End of preparation step
		
		barrier(CLK_LOCAL_MEM_FENCE);
		
		offset =  mad24(groupId, 1024, ii);
		in += offset;
		out += offset;
		a[0] = in[0];
		a[1] = in[128];
		a[2] = in[256];
		a[3] = in[384];
		a[4] = in[512];
		a[5] = in[640];
		a[6] = in[768];
		a[7] = in[896];
		fftKernel8(a+0, dir);
		angf = (float) ii;
		ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[4] = complexMul(a[4], w);
		ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[5] = complexMul(a[5], w);
		ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[6] = complexMul(a[6], w);
		ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[7] = complexMul(a[7], w);
		lMemStore = sMem + ii;
		j = ii & 7;
		i = ii >> 3;
		lMemLoad = sMem + mad24(j, 130, i);
		lMemStore[0] = a[0].x;
		lMemStore[130] = a[1].x;
		lMemStore[260] = a[2].x;
		lMemStore[390] = a[3].x;
		lMemStore[520] = a[4].x;
		lMemStore[650] = a[5].x;
		lMemStore[780] = a[6].x;
		lMemStore[910] = a[7].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		a[4].x = lMemLoad[64];
		a[5].x = lMemLoad[80];
		a[6].x = lMemLoad[96];
		a[7].x = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[130] = a[1].y;
		lMemStore[260] = a[2].y;
		lMemStore[390] = a[3].y;
		lMemStore[520] = a[4].y;
		lMemStore[650] = a[5].y;
		lMemStore[780] = a[6].y;
		lMemStore[910] = a[7].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		a[4].y = lMemLoad[64];
		a[5].y = lMemLoad[80];
		a[6].y = lMemLoad[96];
		a[7].y = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel8(a+0, dir);
		angf = (float) (ii >> 3);
		ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[4] = complexMul(a[4], w);
		ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[5] = complexMul(a[5], w);
		ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[6] = complexMul(a[6], w);
		ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[7] = complexMul(a[7], w);
		lMemStore = sMem + ii;
		j = (ii & 63) >> 3;
		i = mad24(ii >> 6, 8, ii & 7);
		lMemLoad = sMem + mad24(j, 136, i);
		lMemStore[0] = a[0].x;
		lMemStore[136] = a[1].x;
		lMemStore[272] = a[2].x;
		lMemStore[408] = a[3].x;
		lMemStore[544] = a[4].x;
		lMemStore[680] = a[5].x;
		lMemStore[816] = a[6].x;
		lMemStore[952] = a[7].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[32];
		a[2].x = lMemLoad[64];
		a[3].x = lMemLoad[96];
		a[4].x = lMemLoad[16];
		a[5].x = lMemLoad[48];
		a[6].x = lMemLoad[80];
		a[7].x = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[136] = a[1].y;
		lMemStore[272] = a[2].y;
		lMemStore[408] = a[3].y;
		lMemStore[544] = a[4].y;
		lMemStore[680] = a[5].y;
		lMemStore[816] = a[6].y;
		lMemStore[952] = a[7].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[32];
		a[2].y = lMemLoad[64];
		a[3].y = lMemLoad[96];
		a[4].y = lMemLoad[16];
		a[5].y = lMemLoad[48];
		a[6].y = lMemLoad[80];
		a[7].y = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		fftKernel4(a+4, dir);
		angf = (float) (ii >> 6);
		ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		angf = (float) ((128 + ii) >>6);
		ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[5] = complexMul(a[5], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[6] = complexMul(a[6], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[7] = complexMul(a[7], w);
		lMemStore = sMem + ii;
		j = ii >> 6;
		i = ii & 63;
		lMemLoad = sMem + mad24(j, 256, i);
		lMemStore[0] = a[0].x;
		lMemStore[256] = a[1].x;
		lMemStore[512] = a[2].x;
		lMemStore[768] = a[3].x;
		lMemStore[128] = a[4].x;
		lMemStore[384] = a[5].x;
		lMemStore[640] = a[6].x;
		lMemStore[896] = a[7].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[64];
		a[2].x = lMemLoad[128];
		a[3].x = lMemLoad[192];
		a[4].x = lMemLoad[512];
		a[5].x = lMemLoad[576];
		a[6].x = lMemLoad[640];
		a[7].x = lMemLoad[704];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[256] = a[1].y;
		lMemStore[512] = a[2].y;
		lMemStore[768] = a[3].y;
		lMemStore[128] = a[4].y;
		lMemStore[384] = a[5].y;
		lMemStore[640] = a[6].y;
		lMemStore[896] = a[7].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[64];
		a[2].y = lMemLoad[128];
		a[3].y = lMemLoad[192];
		a[4].y = lMemLoad[512];
		a[5].y = lMemLoad[576];
		a[6].y = lMemLoad[640];
		a[7].y = lMemLoad[704];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		fftKernel4(a+4, dir);
		out[0] = a[0];
		out[128] = a[4];
		out[256] = a[1];
		out[384] = a[5];
		out[512] = a[2];
		out[640] = a[6];
		out[768] = a[3];
		out[896] = a[7];
	
		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

		// Do finalization step to convert to Speex FFT format
		// There are one thread for each four positions in the FFT length
		// Not adapted to multi-channel filter, only mono

		int workItemOffset2 = lId * 4 + groupOffset;
		
		// The first thread for each FFT length
		if(lId == 0)
		{
			E[groupOffset + 0] = out_orig[workItemOffset2 + 0].x;
			E[groupOffset + 1] = out_orig[workItemOffset2 + 1].x;
			for(int t=1;t<4;t++)
			{
				E[groupOffset + 2*t] = out_orig[workItemOffset2 + t].y;
				E[groupOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
			}
		}
		else
		{
			for(int t=0;t<4;t++)
			{
				E[workItemOffset + 2*t] = out_orig[workItemOffset2 + t].y;
				E[workItemOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
			}
		}
		
		// End of finalization step 
		
	}
	// Set to value and do FFT
	// Equiv: mdf_opencl_set_array_to_float_value(ocl_y, chan*st->window_size, 0, st->frame_size);
	// 	  speex_echo_opencl_fft_execute2(ocl_y, chan*st->window_size, ocl_Y, chan*st->window_size, 1, commandQueue);
	else
	{
		// Set to zero
		y[groupOffset + 4*lId] = 0.0f;
		y[groupOffset + 4*lId + 1] = 0.0f;
		y[groupOffset + 4*lId + 2] = 0.0f;
		y[groupOffset + 4*lId + 3] = 0.0f;
		
		// Do FFT
		int dir = -1;

		barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
		
		// Do preparation step to convert from Speex FFT format
		// There is one thread for each four positions in the FFT length	
		for(int t=0;t<8;t++)
		{
			in2[workItemOffset + t].x = y[workItemOffset + t] * m;
			in2[workItemOffset + t].y = 0.0f;
		}

		barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
		
		offset =  mad24(groupId, 1024, ii);
		in2 += offset;
		out2 += offset;
		a[0] = in2[0];
		a[1] = in2[128];
		a[2] = in2[256];
		a[3] = in2[384];
		a[4] = in2[512];
		a[5] = in2[640];
		a[6] = in2[768];
		a[7] = in2[896];
		fftKernel8(a+0, dir);
		angf = (float) ii;
		ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[4] = complexMul(a[4], w);
		ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[5] = complexMul(a[5], w);
		ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[6] = complexMul(a[6], w);
		ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[7] = complexMul(a[7], w);
		lMemStore = sMem + ii;
		j = ii & 7;
		i = ii >> 3;
		lMemLoad = sMem + mad24(j, 130, i);
		lMemStore[0] = a[0].x;
		lMemStore[130] = a[1].x;
		lMemStore[260] = a[2].x;
		lMemStore[390] = a[3].x;
		lMemStore[520] = a[4].x;
		lMemStore[650] = a[5].x;
		lMemStore[780] = a[6].x;
		lMemStore[910] = a[7].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		a[4].x = lMemLoad[64];
		a[5].x = lMemLoad[80];
		a[6].x = lMemLoad[96];
		a[7].x = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[130] = a[1].y;
		lMemStore[260] = a[2].y;
		lMemStore[390] = a[3].y;
		lMemStore[520] = a[4].y;
		lMemStore[650] = a[5].y;
		lMemStore[780] = a[6].y;
		lMemStore[910] = a[7].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		a[4].y = lMemLoad[64];
		a[5].y = lMemLoad[80];
		a[6].y = lMemLoad[96];
		a[7].y = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel8(a+0, dir);
		angf = (float) (ii >> 3);
		ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[4] = complexMul(a[4], w);
		ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[5] = complexMul(a[5], w);
		ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[6] = complexMul(a[6], w);
		ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[7] = complexMul(a[7], w);
		lMemStore = sMem + ii;
		j = (ii & 63) >> 3;
		i = mad24(ii >> 6, 8, ii & 7);
		lMemLoad = sMem + mad24(j, 136, i);
		lMemStore[0] = a[0].x;
		lMemStore[136] = a[1].x;
		lMemStore[272] = a[2].x;
		lMemStore[408] = a[3].x;
		lMemStore[544] = a[4].x;
		lMemStore[680] = a[5].x;
		lMemStore[816] = a[6].x;
		lMemStore[952] = a[7].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[32];
		a[2].x = lMemLoad[64];
		a[3].x = lMemLoad[96];
		a[4].x = lMemLoad[16];
		a[5].x = lMemLoad[48];
		a[6].x = lMemLoad[80];
		a[7].x = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[136] = a[1].y;
		lMemStore[272] = a[2].y;
		lMemStore[408] = a[3].y;
		lMemStore[544] = a[4].y;
		lMemStore[680] = a[5].y;
		lMemStore[816] = a[6].y;
		lMemStore[952] = a[7].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[32];
		a[2].y = lMemLoad[64];
		a[3].y = lMemLoad[96];
		a[4].y = lMemLoad[16];
		a[5].y = lMemLoad[48];
		a[6].y = lMemLoad[80];
		a[7].y = lMemLoad[112];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		fftKernel4(a+4, dir);
		angf = (float) (ii >> 6);
		ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		angf = (float) ((128 + ii) >>6);
		ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[5] = complexMul(a[5], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[6] = complexMul(a[6], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[7] = complexMul(a[7], w);
		lMemStore = sMem + ii;
		j = ii >> 6;
		i = ii & 63;
		lMemLoad = sMem + mad24(j, 256, i);
		lMemStore[0] = a[0].x;
		lMemStore[256] = a[1].x;
		lMemStore[512] = a[2].x;
		lMemStore[768] = a[3].x;
		lMemStore[128] = a[4].x;
		lMemStore[384] = a[5].x;
		lMemStore[640] = a[6].x;
		lMemStore[896] = a[7].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[64];
		a[2].x = lMemLoad[128];
		a[3].x = lMemLoad[192];
		a[4].x = lMemLoad[512];
		a[5].x = lMemLoad[576];
		a[6].x = lMemLoad[640];
		a[7].x = lMemLoad[704];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[256] = a[1].y;
		lMemStore[512] = a[2].y;
		lMemStore[768] = a[3].y;
		lMemStore[128] = a[4].y;
		lMemStore[384] = a[5].y;
		lMemStore[640] = a[6].y;
		lMemStore[896] = a[7].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[64];
		a[2].y = lMemLoad[128];
		a[3].y = lMemLoad[192];
		a[4].y = lMemLoad[512];
		a[5].y = lMemLoad[576];
		a[6].y = lMemLoad[640];
		a[7].y = lMemLoad[704];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		fftKernel4(a+4, dir);
		out2[0] = a[0];
		out2[128] = a[4];
		out2[256] = a[1];
		out2[384] = a[5];
		out2[512] = a[2];
		out2[640] = a[6];
		out2[768] = a[3];
		out2[896] = a[7];
		
		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
		
		// Do finalization step to convert to Speex FFT format
		// There are one thread for each four positions in the FFT length

		int workItemOffset2 = lId * 4 + groupOffset;

		// The first thread for each FFT length
		if(lId == 0)
		{
			Y[groupOffset + 0] = out2_orig[workItemOffset2 + 0].x;
			Y[groupOffset + 1] = out2_orig[workItemOffset2 + 1].x;
			for(int t=1;t<4;t++)
			{
				Y[groupOffset + 2*t] = out2_orig[workItemOffset2 + t].y;
				Y[groupOffset + 2*t + 1] = out2_orig[workItemOffset2 + t + 1].x;	
			}
		}
		else
		{
			for(int t=0;t<4;t++)
			{
				Y[workItemOffset + 2*t] = out2_orig[workItemOffset2 + t].y;
				Y[workItemOffset + 2*t + 1] = out2_orig[workItemOffset2 + t + 1].x;	
			}
		}

		// End of finalization step
	}
	
}

/**
 * Specialized kernel for doing a forward FFT and then an inner product
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 1024.
 */ 
__kernel void clFFT_inner_prod_combined(__global float *data_time_domain, __global float2 *in, __global float2 *out, __global float *data_freq_domain, __global float *Xf, __global float *temp_reduction, int frame_size, int S)
{
	
	__local float sMem[1088];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	int lId = get_local_id(0);
	int groupId = get_group_id(0);
	int groupOffset = groupId * 1024;	// Group offset for float
	int groupOffset2 = groupId * 512;	// Group offset for float2
	unsigned int xi = get_local_id(0)*4 + frame_size + groupOffset;	// Used for inner product
	unsigned int yi = get_local_id(0)*4 + frame_size + groupOffset;	// Used for inner product
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[8];
	const float m = 1.0f / 1024.0f;
	ii = lId;
	jj = 0;

	// Do FFT
	int dir = -1;
	
	offset =  mad24(groupId, 1024, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	
	int workItemOffset = lId * 8 + groupOffset;
	
	barrier(CLK_LOCAL_MEM_FENCE);

	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length	
	for(int t=0;t<8;t++)
	{
		in[workItemOffset + t].x = data_time_domain[workItemOffset + t] * m;
		in[workItemOffset + t].y = 0.0f;
	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	
	// End of preparation step
	offset =  mad24(groupId, 1024, ii);
	in += offset;
	out += offset;
	a[0] = in[0];
	a[1] = in[128];
	a[2] = in[256];
	a[3] = in[384];
	a[4] = in[512];
	a[5] = in[640];
	a[6] = in[768];
	a[7] = in[896];
	fftKernel8(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii & 7;
	i = ii >> 3;
	lMemLoad = sMem + mad24(j, 130, i);
	lMemStore[0] = a[0].x;
	lMemStore[130] = a[1].x;
	lMemStore[260] = a[2].x;
	lMemStore[390] = a[3].x;
	lMemStore[520] = a[4].x;
	lMemStore[650] = a[5].x;
	lMemStore[780] = a[6].x;
	lMemStore[910] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	a[4].x = lMemLoad[64];
	a[5].x = lMemLoad[80];
	a[6].x = lMemLoad[96];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[130] = a[1].y;
	lMemStore[260] = a[2].y;
	lMemStore[390] = a[3].y;
	lMemStore[520] = a[4].y;
	lMemStore[650] = a[5].y;
	lMemStore[780] = a[6].y;
	lMemStore[910] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	a[4].y = lMemLoad[64];
	a[5].y = lMemLoad[80];
	a[6].y = lMemLoad[96];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel8(a+0, dir);
	angf = (float) (ii >> 3);
	ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = (ii & 63) >> 3;
	i = mad24(ii >> 6, 8, ii & 7);
	lMemLoad = sMem + mad24(j, 136, i);
	lMemStore[0] = a[0].x;
	lMemStore[136] = a[1].x;
	lMemStore[272] = a[2].x;
	lMemStore[408] = a[3].x;
	lMemStore[544] = a[4].x;
	lMemStore[680] = a[5].x;
	lMemStore[816] = a[6].x;
	lMemStore[952] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[32];
	a[2].x = lMemLoad[64];
	a[3].x = lMemLoad[96];
	a[4].x = lMemLoad[16];
	a[5].x = lMemLoad[48];
	a[6].x = lMemLoad[80];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[136] = a[1].y;
	lMemStore[272] = a[2].y;
	lMemStore[408] = a[3].y;
	lMemStore[544] = a[4].y;
	lMemStore[680] = a[5].y;
	lMemStore[816] = a[6].y;
	lMemStore[952] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[32];
	a[2].y = lMemLoad[64];
	a[3].y = lMemLoad[96];
	a[4].y = lMemLoad[16];
	a[5].y = lMemLoad[48];
	a[6].y = lMemLoad[80];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	angf = (float) (ii >> 6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	angf = (float) ((128 + ii) >>6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii >> 6;
	i = ii & 63;
	lMemLoad = sMem + mad24(j, 256, i);
	lMemStore[0] = a[0].x;
	lMemStore[256] = a[1].x;
	lMemStore[512] = a[2].x;
	lMemStore[768] = a[3].x;
	lMemStore[128] = a[4].x;
	lMemStore[384] = a[5].x;
	lMemStore[640] = a[6].x;
	lMemStore[896] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[64];
	a[2].x = lMemLoad[128];
	a[3].x = lMemLoad[192];
	a[4].x = lMemLoad[512];
	a[5].x = lMemLoad[576];
	a[6].x = lMemLoad[640];
	a[7].x = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[256] = a[1].y;
	lMemStore[512] = a[2].y;
	lMemStore[768] = a[3].y;
	lMemStore[128] = a[4].y;
	lMemStore[384] = a[5].y;
	lMemStore[640] = a[6].y;
	lMemStore[896] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[64];
	a[2].y = lMemLoad[128];
	a[3].y = lMemLoad[192];
	a[4].y = lMemLoad[512];
	a[5].y = lMemLoad[576];
	a[6].y = lMemLoad[640];
	a[7].y = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	out[0] = a[0];
	out[128] = a[4];
	out[256] = a[1];
	out[384] = a[5];
	out[512] = a[2];
	out[640] = a[6];
	out[768] = a[3];
	out[896] = a[7];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length

	int workItemOffset2 = lId * 4 + groupOffset;

	// The first thread for each FFT length
	if(lId == 0)
	{
		data_freq_domain[groupOffset + 0] = out_orig[workItemOffset2 +  0].x;
		data_freq_domain[groupOffset + 1] = out_orig[workItemOffset2 + 1].x;
		for(int t=1;t<4;t++)
		{
			data_freq_domain[groupOffset + 2*t] = out_orig[workItemOffset2 + t].y;
			data_freq_domain[groupOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
		}
	}
	else
	{
		for(int t=0;t<4;t++)
		{
			data_freq_domain[workItemOffset + 2*t] = out_orig[workItemOffset2 + t].y;
			data_freq_domain[workItemOffset + 2*t + 1] = out_orig[workItemOffset2 + t + 1].x;	
		}
	}

	// End of finalization step

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Calculate inner product (last reduction is done on host)
	if(get_local_id(0) < 128) {
		sMem[lId] = 0.0f;
		//sMem[lId] = fma(data_time_domain[xi], data_time_domain[yi], sMem[lId]);	
		//sMem[lId] = fma(data_time_domain[xi+1], data_time_domain[yi+1], sMem[lId]);
		//sMem[lId] += 1.0f;
		sMem[lId] += data_time_domain[xi] * data_time_domain[yi];	
		sMem[lId] += data_time_domain[xi+1] * data_time_domain[yi+1];
		sMem[lId] += data_time_domain[xi+2] * data_time_domain[yi+2];
		sMem[lId] += data_time_domain[xi+3] * data_time_domain[yi+3];
	}
	else
		sMem[lId] = 0.0f;

	barrier(CLK_LOCAL_MEM_FENCE);

	// do reduction in shared mem
	for(unsigned int s = get_local_size(0) / 2; s > 0; s >>= 1) 
	{
		if(lId < s)
		{
			sMem[lId] += sMem[lId + s];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	// write result for this block to global mem
	if (lId == 0) 
	{
		temp_reduction[get_group_id(0)] = sMem[0];
	}

	//barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
	
	// Power spectrum accum - must do twice the normal work, usually run with 128 threads
	// Loop through output channels, because of lack of atomic adds 
// 	if(get_group_id(0))
// 	{	
// 		j = get_local_id(0);
// 		k = j * 4 - 1;
// 		l = j * 4 + 1;
// 		int speak_offset = 0;
// 
// 		if(lId == 0)
// 		{
// 			for(int s=0;s<get_num_groups(0);s++)
// 			{
// 				Xf[0] += data_freq_domain[speak_offset + 0] * data_freq_domain[speak_offset + 0];
// 				Xf[1] += data_freq_domain[speak_offset + 1] * data_freq_domain[speak_offset + 1] + data_freq_domain[speak_offset + 2] * data_freq_domain[speak_offset + 2];
// 				speak_offset += 2*frame_size;
// 			}
// 		}
// 		else
// 		{
// 			for(int s=0;s<get_num_groups(0);s++)
// 			{
// 				Xf[2*j] += data_freq_domain[speak_offset + k] * data_freq_domain[speak_offset + k] + data_freq_domain[speak_offset + k + 1] * data_freq_domain[speak_offset + k + 1];
// 				Xf[2*j + 1] += data_freq_domain[speak_offset + l] * data_freq_domain[speak_offset + l] + data_freq_domain[speak_offset + l + 1] * data_freq_domain[speak_offset + l + 1];
// 				speak_offset += 2*frame_size;
// 			}
// 		}
// 		// Set the last element (frame_size + 1)
// 		if(lId == (get_local_size(0) - 1))
// 		{
// 			for(int s=0;s<get_num_groups(0);s++)
// 			{
// 				Xf[frame_size] += data_freq_domain[speak_offset + (frame_size*2) - 1] * data_freq_domain[speak_offset + (frame_size*2) - 1];
// 				speak_offset += 2*frame_size;
// 			}
// 		}
// 	}
	
}

/**
 * Regular inverse FFT, combined with the spectral mul accum (long) reduce kernel
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060, 
 * for 1D input of length 1024.
 */ 
__kernel void clFFT_inverse_spectral_mul_accum_reduce(__global float *Y, __global float2 *in, __global float2 *out, __global float *e, __global float *temp_acc, int M, int N, int S)
{
	
	__local float sMem[1088];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	float2 w;
	float ang, angf, ang1;
	int dir = 1;	// Inverse
	__local float *lMemStore, *lMemLoad;
	float2 a[8];
	int lId = get_local_id(0);
	int groupId = get_group_id(0);
	const float m = 1.0f / 1024.0f;
	int groupOffset = groupId * 1024;
	ii = lId;
	jj = 0;

	offset =  mad24(groupId, 1024, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	int workItemOffset = lId * 8 + groupOffset;
	

	// Do the spectral mul accum reduce, 8 elementents for each thread
	for(int t=0;t<8;t++)
	{
		Y[groupOffset + lId*8 + t] = 0.0f;	
	}

	for(int c=0;c<M;c++)
	{
		for(int t=0;t<8;t++)
		{
			Y[groupOffset + lId*2 + t] += temp_acc[groupId*M*N + c*N + lId*8 + t];
		}
	}

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	if(lId == 0)
	{		
		in[groupOffset + 0].x = Y[groupOffset + 0];
		in[groupOffset + 0].y = 0.0f;
		for(int t=1;t<4;t++)
		{
			in[groupOffset + t].x = Y[groupOffset + (2*t - 1)];
			in[groupOffset + t].y = Y[groupOffset + (2*t - 1) + 1];
		}
		in[groupOffset + 512].x = Y[groupOffset + 1024 - 1];
		in[groupOffset + 512].y = 0.0f;
	}
	else
	{
		in[groupOffset + 4*lId].x = Y[groupOffset + lId*8 - 1];
		in[groupOffset + 4*lId].y = Y[groupOffset + lId*8];
		in[groupOffset + 4*lId + 1].x = Y[groupOffset + lId*8 + 1];
		in[groupOffset + 4*lId + 1].y = Y[groupOffset + lId*8 + 2];
		
		for(int t=0;t<4;t++)
		{
			in[groupOffset + 4*lId + t].x = Y[groupOffset + lId*8 - 1 + 2*t];
			in[groupOffset + 4*lId + t].y = Y[groupOffset + lId*8 + 2*t];
		}
	}

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	// Generate negative frequencies for symmetric input 
	// Source: http://www.dspguide.com/ch12/1.htm
	// For the remaining n/2 threads that needs to mirror
	//for(int t=0;t<8;t++)
	//{
	//	in[groupOffset + 1023 - lId*8 - t].x = in[groupOffset + lId*8 + t + 1].x;
	//	in[groupOffset + 1023 - lId*8 - t].y = -in[groupOffset + lId*8 + t + 1].y;
	//}
	for(int t=1;t<=4;t++)
	{
		in[groupOffset + 1024 - lId*4 - t].x = in[groupOffset + lId*4 + t].x;
		in[groupOffset + 1024 - lId*4 - t].y = -in[groupOffset + lId*4 + t].y;
	}

	barrier(CLK_LOCAL_MEM_FENCE);

	in += offset;
	out += offset;
	a[0] = in[0];
	a[1] = in[128];
	a[2] = in[256];
	a[3] = in[384];
	a[4] = in[512];
	a[5] = in[640];
	a[6] = in[768];
	a[7] = in[896];
	fftKernel8(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 1024.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii & 7;
	i = ii >> 3;
	lMemLoad = sMem + mad24(j, 130, i);
	lMemStore[0] = a[0].x;
	lMemStore[130] = a[1].x;
	lMemStore[260] = a[2].x;
	lMemStore[390] = a[3].x;
	lMemStore[520] = a[4].x;
	lMemStore[650] = a[5].x;
	lMemStore[780] = a[6].x;
	lMemStore[910] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	a[4].x = lMemLoad[64];
	a[5].x = lMemLoad[80];
	a[6].x = lMemLoad[96];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[130] = a[1].y;
	lMemStore[260] = a[2].y;
	lMemStore[390] = a[3].y;
	lMemStore[520] = a[4].y;
	lMemStore[650] = a[5].y;
	lMemStore[780] = a[6].y;
	lMemStore[910] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	a[4].y = lMemLoad[64];
	a[5].y = lMemLoad[80];
	a[6].y = lMemLoad[96];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel8(a+0, dir);
	angf = (float) (ii >> 3);
	ang = dir * ( 2.0f * M_PI * 1.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	ang = dir * ( 2.0f * M_PI * 4.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[4] = complexMul(a[4], w);
	ang = dir * ( 2.0f * M_PI * 5.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 6.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 7.0f / 128.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = (ii & 63) >> 3;
	i = mad24(ii >> 6, 8, ii & 7);
	lMemLoad = sMem + mad24(j, 136, i);
	lMemStore[0] = a[0].x;
	lMemStore[136] = a[1].x;
	lMemStore[272] = a[2].x;
	lMemStore[408] = a[3].x;
	lMemStore[544] = a[4].x;
	lMemStore[680] = a[5].x;
	lMemStore[816] = a[6].x;
	lMemStore[952] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[32];
	a[2].x = lMemLoad[64];
	a[3].x = lMemLoad[96];
	a[4].x = lMemLoad[16];
	a[5].x = lMemLoad[48];
	a[6].x = lMemLoad[80];
	a[7].x = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[136] = a[1].y;
	lMemStore[272] = a[2].y;
	lMemStore[408] = a[3].y;
	lMemStore[544] = a[4].y;
	lMemStore[680] = a[5].y;
	lMemStore[816] = a[6].y;
	lMemStore[952] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[32];
	a[2].y = lMemLoad[64];
	a[3].y = lMemLoad[96];
	a[4].y = lMemLoad[16];
	a[5].y = lMemLoad[48];
	a[6].y = lMemLoad[80];
	a[7].y = lMemLoad[112];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	angf = (float) (ii >> 6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	angf = (float) ((128 + ii) >>6);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[5] = complexMul(a[5], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[6] = complexMul(a[6], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[7] = complexMul(a[7], w);
	lMemStore = sMem + ii;
	j = ii >> 6;
	i = ii & 63;
	lMemLoad = sMem + mad24(j, 256, i);
	lMemStore[0] = a[0].x;
	lMemStore[256] = a[1].x;
	lMemStore[512] = a[2].x;
	lMemStore[768] = a[3].x;
	lMemStore[128] = a[4].x;
	lMemStore[384] = a[5].x;
	lMemStore[640] = a[6].x;
	lMemStore[896] = a[7].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[64];
	a[2].x = lMemLoad[128];
	a[3].x = lMemLoad[192];
	a[4].x = lMemLoad[512];
	a[5].x = lMemLoad[576];
	a[6].x = lMemLoad[640];
	a[7].x = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[256] = a[1].y;
	lMemStore[512] = a[2].y;
	lMemStore[768] = a[3].y;
	lMemStore[128] = a[4].y;
	lMemStore[384] = a[5].y;
	lMemStore[640] = a[6].y;
	lMemStore[896] = a[7].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[64];
	a[2].y = lMemLoad[128];
	a[3].y = lMemLoad[192];
	a[4].y = lMemLoad[512];
	a[5].y = lMemLoad[576];
	a[6].y = lMemLoad[640];
	a[7].y = lMemLoad[704];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	fftKernel4(a+4, dir);
	out[0] = a[0];
	out[128] = a[4];
	out[256] = a[1];
	out[384] = a[5];
	out[512] = a[2];
	out[640] = a[6];
	out[768] = a[3];
	out[896] = a[7];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length

	// clFFT_Inverse
	for(int t=0;t<8;t++)
	{
		e[workItemOffset + t] = out_orig[workItemOffset + t].x;
	}
	
	// End of finalization step
	
}
