#define WORD2INT(x) ((x) < -32767.5f ? (int)-32768 : ((x) > 32766.5f ? (int)32767 : floor(0.5f+(x))))
#pragma OPENCL EXTENSION cl_khr_byte_addressable_store : enable 

// Globalsize: window_size * M  
//__kernel void clMDF_prepare_shift_preemph_freq_data(__global float *input, __global float *x, __global float *X, __global float *tempX, __global short *far_end, __global float *memD, __global float *memX, float preemph, unsigned int frame_size, unsigned int window_size, unsigned int num_channels, unsigned int num_speakers, __local int* tmpMemX, __local float *tmpMemD)
__kernel void clMDF_prepare_shift_preemph_freq_data(__global float *input, __global float *x, __global float *X, __global float *tempX, __global short *far_end, __global float *memD, __global float *memX, float preemph, unsigned int frame_size, unsigned int window_size, unsigned int num_channels, unsigned int num_speakers, __local int* tmpMemX, __local float *tmpMemD)
{
	
	unsigned int i = get_global_id(0);	// Goes to window_size (N)
	unsigned int j = get_global_id(1);	// Goes to M
	unsigned int k = get_global_id(2);	// Goes to K
	unsigned int speak = get_global_id(2);	// Goes to K
	unsigned int lsize = get_local_size(0);
	unsigned int frame_ratio = frame_size / lsize;
	unsigned int tid = get_local_id(0);
	
	unsigned int chan;	// Usually just one channel, but loop if it not the case
	
	// If I is witin frame size, do prepare and shift_preemph  
	// Done on a single block, parallelize?
	if(i < lsize && j == 0 && k == 0)
	{
		
		// clMDF_prepare
		if(i == 0)
		{
			for(chan = 0; chan < num_channels; chan++)
			{
				tmpMemD[chan*frame_size + frame_ratio*tid + 0] = memD[chan];
				for(int s = 1;s<frame_ratio;s++)
					tmpMemD[lsize*chan*frame_ratio + frame_ratio*tid + s] = input[chan*frame_size + frame_ratio*i + s - 1];
			}
		}
		else
		{
			for(chan = 0; chan < num_channels; chan++)
			{
				for(int s = 0;s<frame_ratio;s++)
					tmpMemD[lsize*chan*frame_ratio + frame_ratio*tid + s] = input[chan*frame_size + frame_ratio*i + s - 1];
			}
		}
		/*if(i == (frame_size/2 - 1))
		{
			for(chan = 0; chan < num_channels; chan++)
			{
				memD[chan] = input[chan * frame_size + frame_size - 1];	
				
			}
		}*/
	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	
	if(i < lsize && j == 0 && k == 0)
	{
		for(chan = 0; chan < num_channels; chan++)
		{
			//memD[chan] = input[chan * frame_size + frame_size - 1];
			for(int s = 0;s<frame_ratio;s++)
			{
				input[chan * frame_size + frame_ratio*i + s] = input[chan*frame_size + frame_ratio*i + s] - (preemph * tmpMemD[lsize*chan*frame_ratio + frame_ratio*tid + s]);
			}
			
		}

	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
		
	if(i < frame_size/2 && j == 0)
	{
		// clMDF_shift_preemph
		// Temporary value in place of memX (updated on each iteration in sequential code)
		// Notice that we only store the values needed for update, not the shift
		// This is done to save local mem. (only frame_size*K, instead of window_size)
		if(i == 0)
		{
			//tmpMemX[speak*window_size] = memX[speak];
			//tmpMemX[speak*window_size + 1] = far_end[speak];
			tmpMemX[speak*frame_size] = memX[speak];
			tmpMemX[speak*frame_size + 1] = far_end[speak];
			memX[speak] = far_end[(frame_size - 1)*num_speakers + speak];   
		}
		else
		{				
			tmpMemX[speak*frame_size + 2*i] = far_end[((i*2-1) * num_speakers) + speak];
			tmpMemX[speak*frame_size + 2*i + 1] = far_end[((i*2) * num_speakers) + speak];
		}
	}
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	if(i < frame_size/2 && j == 0)
	{
		
		// Shift data from the last frame in the time domain data
		x[speak*window_size + i*2] = x[speak*window_size + 2*i + frame_size];
		x[speak*window_size + i*2+1] = x[speak*window_size + 2*i + 1 + frame_size];
		//x[speak*window_size + i*2] = 5000.0f;
		//x[speak*window_size + i*2+1] = 6000.0f;

		// Insert new data in the new frame and apply preemphasis
		x[speak*window_size + i*2 + frame_size] = far_end[(2*i)*num_speakers + speak] - (preemph * tmpMemX[speak*frame_size + i*2]);
		x[speak*window_size + i*2 + 1 + frame_size] = far_end[(2*i + 1)*num_speakers + speak] - (preemph * tmpMemX[speak*frame_size + i*2 + 1]);
		//x[speak*window_size + i*2 + frame_size] = 8000.0f;
		//x[speak*window_size + i*2 + 1 + frame_size] = 9000.0f;
	}
		
	// clMDF_shift_freq_data
	for(speak = 0; speak < num_speakers; speak++) {
		X[(j+1)*window_size*num_speakers+speak*window_size+i] = tempX[j*window_size*num_speakers+speak*window_size+i];
	}

	//if(get_global_id(0) == 0 && get_global_id(1) == 0) 
	//{
		//for(speak = 0; speak < num_speakers; speak++)
		//{
			//memX[speak] = far_end[(frame_size - 1)*num_speakers + speak];
		//}
	//}
}

/**
 * Based on reduction kernel from the NVIDIA CUDA SDK (reduce2 kernel)
 * Optional (naive) support for batch runs, batch_size is the number of elements in 
 * the batch, batch_pitch is the distance between elements in the batch (must be the same for x and y).
*/ 
__kernel void clMDF_inner_product_reduce(__global float *x, unsigned int x_offset, __global float *y, unsigned int y_offset, __global float *g_odata, unsigned int n, unsigned int batch_size, unsigned int batch_pitch, __local float* sdata)
{
	// load shared mem
	unsigned int tid = get_local_id(0);
	unsigned int i = get_global_id(0);
	unsigned int xi = i*2 + x_offset;
	unsigned int yi = i*2 + y_offset;

	if(i < n) {
		sdata[tid] = 0.0f;
		
		for(int k=0;k<batch_size;k++)
		{
			sdata[tid] += x[k*batch_pitch + xi] * y[k*batch_pitch + yi];	
			sdata[tid] += x[k*batch_pitch + xi + 1] * y[k*batch_pitch + yi + 1];
		}
	}
	else
		sdata[tid] = 0.0f;

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// do reduction in shared mem
	for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
	{
		if (tid < s)
		{
			sdata[tid] += sdata[tid + s];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	// write result for this block to global mem
	if (tid == 0) 
		g_odata[get_group_id(0)] = sdata[0];
}

/**
 * Combined kernel with 3*power_spectrum_accum and 1*inner_product_reduce 
*/
__kernel void clMDF_power_spectrum_inner_product(
	__global float *x, __global float *E, __global float *Y, __global float *X,
        __global float *Rf, __global float *Yf, __global float *Xf,
	// Offsets to power pectrum accum
        unsigned int E_offset, unsigned int Y_offset, unsigned int X_offset, 
	// Offsets to inner product
	unsigned int x1_offset, unsigned int x2_offset,
        unsigned int C, unsigned int K, unsigned int window_size,
        unsigned int N_power_spectrum,
	unsigned int n_inner_product,
	__global float *temp_odata,
	__local float* ldata)
{
	
	unsigned int j = get_global_id(0);
        unsigned int i = j * 2 - 1;

	unsigned int tid = get_local_id(0);
	unsigned int xi = j*2 + x1_offset;
	unsigned int yi = j*2 + x2_offset;
	
	// Optimize for regular case (mono mic and speaker)
	if(C == 0 && K == 0)
	{
		if(j == 0)
		{
			Rf[j] += E[0] * E[0];
			Yf[j] += Y[0] * Y[0];
			Xf[j] += X[0] * X[0];
		}
		else if(i < (N_power_spectrum-1))
		{
			Rf[j] += E[i] * E[i] + E[i+1] * E[i+1];
			Yf[j] += Y[i] * Y[i] + Y[i+1] * Y[i+1];
			Xf[j] += X[i] * X[i] + X[i+1] * X[i+1];
		}
		if(j == (get_global_size(0) - 1))
		{
			i += 2;
			Rf[j+1] += E[i] * E[i];
			Yf[j+1] += Y[i] * Y[i];
			Xf[j+1] += X[i] * X[i];
		}
	}
	// In case of multi-channel (rewrite for better performance if needed)
	else
	{
		if(j == 0)
		{
			unsigned int chan_offset, speak_offset;
			for(unsigned int chan = 0;chan<C;chan++)
			{
				chan_offset = chan*window_size;
				Rf[j] += E[0 + chan_offset] * E[0 + chan_offset];
				Yf[j] += Y[0 + chan_offset] * Y[0 + chan_offset];
			}
			for(unsigned int speak = 0;speak<K;speak++)
			{
				speak_offset = speak*window_size;
				Xf[j] += X[0 + speak_offset] * X[0 + speak_offset];
			}
		}
		else if(i < (N_power_spectrum-1))
		{
			unsigned int chan_offset, speak_offset;
			for(unsigned int chan = 0;chan<C;chan++)
			{
				chan_offset = chan*window_size;
				Rf[j] += E[i + chan_offset] * E[i + chan_offset] + E[i+1 + chan_offset] * E[i+1 + chan_offset];
				Yf[j] += Y[i + chan_offset] * Y[i + chan_offset] + Y[i+1 + chan_offset] * Y[i+1 + chan_offset];
			}
			for(unsigned int speak = 0;speak<K;speak++)
			{
				speak_offset = speak*window_size;
				Xf[j] += X[i + speak_offset] * X[i + speak_offset] + X[i+1 + speak_offset] * X[i+1 + speak_offset];
			}
		}
		if(j == (get_global_size(0) - 1))
		{
			i += 2;
			unsigned int chan_offset, speak_offset;
			for(unsigned int chan = 0;chan<C;chan++)
			{
				chan_offset = chan*window_size;
				Rf[j+1] += E[i + chan_offset] * E[i + chan_offset];
				Yf[j+1] += Y[i + chan_offset] * Y[i + chan_offset];
			}
			for(unsigned int speak = 0;speak<K;speak++)
			{
				speak_offset = speak*window_size;
				Xf[j+1] += X[i + speak_offset] * X[i + speak_offset];
			}
		}
	}

	// Again, optimize for 1 speaker
	if(K == 0)
	{
		if(j < (n_inner_product/2)) {
			ldata[tid] = 0.0f;
			ldata[tid] = fma(x[xi], x[yi], ldata[tid]);	
			ldata[tid] = fma(x[xi+1], x[yi+1], ldata[tid]);
		}
		else
			ldata[tid] = 0.0f;

		barrier(CLK_LOCAL_MEM_FENCE);

		// do reduction in shared mem
		for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
		{
			if (tid < s)
			{
				ldata[tid] += ldata[tid + s];
				//ldata[tid] = 1.0;
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		// write result for this block to global mem
		if (tid == 0) 
			temp_odata[get_group_id(0)] = ldata[0];
	}
	// Several speakers
	else
	{
		unsigned int x_speak_offset;
		for(unsigned int speak = 0;speak<K;speak++)
		{
			x_speak_offset = speak*window_size;
			if(j < (n_inner_product/2)) {
				ldata[tid] = 0.0f;
				ldata[tid] = fma(x[xi + x_speak_offset], x[yi + x_speak_offset], ldata[tid]);	
				ldata[tid] = fma(x[xi+1 + x_speak_offset], x[yi+1 + x_speak_offset], ldata[tid]);
			}
			else
				ldata[tid] = 0.0f;

			barrier(CLK_LOCAL_MEM_FENCE);

			// do reduction in shared mem
			for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
			{
				if (tid < s)
				{
					ldata[tid] += ldata[tid + s];
					//ldata[tid] = 1.0;
				}
				barrier(CLK_LOCAL_MEM_FENCE);
			}
			
			barrier(CLK_LOCAL_MEM_FENCE);

			// write result for this block to global mem
			if (tid == 0) 
				temp_odata[get_group_id(0) + get_num_groups(0)*speak] = ldata[0];

			barrier(CLK_LOCAL_MEM_FENCE);
		}
	}
	
}

// lMem is assumed to be of size N
__kernel void clMDF_spectral_mul_accum(__global float *X, __global float *Y, __global float *acc, __global float *Rf, __global float *Yf, __global float *Xf, unsigned int N, unsigned int num_speakers, unsigned int M)
{
	int i = get_global_id(0);
	unsigned int tid = get_local_id(0);
	unsigned int chan = get_global_id(1);
	unsigned int c;
	unsigned int M_org = M / num_speakers;
	
	// Temporary accum registers
	float tmp1 = 0.0f;
	float tmp2 = 0.0f;
	float tmp3 = 0.0f;

	if(i == 0)
	{
		for(c=0;c<M;c++)
		{
			tmp1 += X[c*N + 0]*Y[chan*N*num_speakers*M_org + c*N + 0];
			tmp2 += (X[c*N + 1]*Y[chan*N*num_speakers*M_org + c*N + 1] - X[c*N + 2]*Y[chan*N*num_speakers*M_org + c*N + 2]);
			//tmp1 += 1.0f;
			//tmp2 += 1.0f;
		}
	}
	else if(i < (get_global_size(0) - 1))
	{
		for(c=0;c<M;c++)
		{
			tmp1 += (X[c*N + 2*i - 1]*Y[chan*N*num_speakers*M_org + c*N + 2*i - 1] - X[c*N + 2*i]*Y[chan*N*num_speakers*M_org + c*N + 2*i]);
			tmp2 += (X[c*N + 2*i]*Y[chan*N*num_speakers*M_org + c*N + 2*i - 1] + X[c*N + 2*i - 1]*Y[chan*N*num_speakers*M_org + c*N + 2*i]);
			//tmp1 += 2.0f;
			//tmp2 += 2.0f; 
		}
	}
	else if(i == (get_global_size(0) - 1))
	{	
		for(c=0;c<M;c++)
		{
			tmp1 += (X[c*N + 2*i - 1]*Y[chan*N*num_speakers*M_org + c*N + 2*i - 1] - X[c*N + 2*i]*Y[chan*N*num_speakers*M_org + c*N + 2*i]);
			tmp2 += (X[c*N + 2*i]*Y[chan*N*num_speakers*M_org + c*N + 2*i - 1] + X[c*N + 2*i - 1]*Y[chan*N*num_speakers*M_org + c*N + 2*i]);
			//tmp2 += (X[c*N + (N-2)]*Y[chan*N*num_speakers*M_org + c*N + (N-3)] + X[c*N + (N-3)]*Y[chan*N*num_speakers*M_org + c*N + (N-2)]);
			tmp3 += X[c*N + (N-1)]*Y[chan*N*num_speakers*M_org + c*N + (N-1)];
			//tmp1 += 3.0f;
			//tmp2 += 3.0f;
		}
	}
	
	// Write out to global memory
	if(i > 0 && i < (get_global_size(0) - 1))
	{
		acc[chan*N + 2*i - 1] = tmp1;
		acc[chan*N + 2*i] = tmp2;
		//acc[chan*N + 2*i - 1] = 9000.0f;
		//acc[chan*N + 2*i] = 8000.0f;
	}
	else if(i == 0)
	{
		acc[chan*N] = tmp1;
		//acc[chan*N + 2*i + 1] = tmp2;
	}
	else if(i == (get_global_size(0) - 1))
	{
		acc[chan*N + 2*i - 1] = tmp1;
		acc[chan*N + 2*i] = tmp2;
		acc[chan*N + 2*i + 1] = tmp3;
	}
	
	if(chan == 0)
	{
		Rf[i] = 0.0f;
		Yf[i] = 0.0f;
		Xf[i] = 0.0f;
		if(i == get_global_size(0) - 1)
		{
			Rf[N/2] = 0.0f;
			Yf[N/2] = 0.0f;
			Xf[N/2] = 0.0f;
		}
	}
}

__kernel void clMDF_compute_foreground_filter(__global float *e, __global float *input, unsigned int window_size, unsigned int frame_size, unsigned int num_channels)
{
	if(get_global_id(1) < num_channels)
	{
		unsigned int i = get_global_id(0);
		unsigned int chan = get_global_id(1);

		e[chan*window_size + i] = input[chan*frame_size + i] - e[chan*window_size + i + frame_size];
	}
}

__kernel void clMDF_adjust_prop_phase1_reduce(__global float *W, __global float *g_odata, unsigned int M, unsigned int window_size, unsigned int n, unsigned int P, __local float* sdata)
{	
	unsigned int j = get_global_id(0);
	unsigned int i = get_global_id(2);
	unsigned int tid = get_local_id(0);

	sdata[tid] = 0.0f;
	// Loop through channels and add to local memory
	for(int p=0;p<P;p++)
	{
		sdata[tid] += W[p*window_size*M + i*window_size+j] * W[p*window_size*M + i*window_size+j];
		barrier(CLK_LOCAL_MEM_FENCE);
	}
		
	// Do local memory reduction
	for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
	{
		if (tid < s)
		{
			sdata[tid] += sdata[tid + s];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	// Write result for this work-group to global mem
	if (tid == 0) 
		g_odata[get_global_id(2)*get_num_groups(0) + get_group_id(0)] = sdata[0];

}

__kernel void clMDF_weighted_spectral_mul_conj(__global float *w, __global float *p, __global float *X, __global float *Y, unsigned int Y_offset, __global float *prod, __global float *W, unsigned int window_size, unsigned int num_channels, unsigned int num_speakers)
{

	unsigned int i = get_global_id(0) * 2 - 1;
	unsigned int j = get_global_id(0);
	unsigned int u = get_global_id(1);
	unsigned int tid = get_global_id(0);
	
	int speak = get_global_id(2) / num_channels;
	int chan = get_global_id(2) - speak*num_channels;
	
	unsigned int X_offset = (u+1)*window_size*num_speakers+speak*window_size;
	Y_offset = chan*window_size;
	unsigned int W_offset = chan * window_size * num_speakers * get_global_size(1) + u * window_size * num_speakers + speak * window_size;
	unsigned int p_offset = u;

	float temp, temp2;
	
	if(tid == 0)
	{
		temp = p[p_offset] * w[0];
		temp *= X[X_offset + 0] * Y[Y_offset + 0];
		W[W_offset] += temp;
		if(get_global_id(1) == 0)
			prod[0] = temp;
	}
	else if(tid <= get_global_size(0) - 1)
	{
		temp = temp2 = p[p_offset] * w[j];
		temp *= ((X[X_offset + i] * Y[Y_offset + i]) + X[X_offset + i + 1] * Y[Y_offset + i+1]);
		temp2 *= ((-X[X_offset + i + 1] * Y[Y_offset + i]) + X[X_offset + i] * Y[Y_offset + i + 1]);
		W[W_offset + i] += temp;
		W[W_offset + i + 1] += temp2;

		if(get_global_id(1) == 0)
		{
			prod[i] = temp;
			prod[i+1] = temp2;
		}
	}
	if(tid == get_global_size(0) - 1)
	{
		i += 2;
		j += 1;
		temp = p[p_offset] * w[j];
		temp *= (X[X_offset + i] * Y[Y_offset + i]);
		W[W_offset + i] += temp;
		if(get_global_id(1) == 0)
			prod[i] = temp;
	}
	
}

__kernel void clMDF_set_array_to_float_value(__global float *input_array, unsigned int input_offset, float value)
{
	
	unsigned int i = get_global_id(0);

	input_array[input_offset + i] = value;
	
}

/**
 * Because of the offset, it is possible to shift data with input=output
 */
__kernel void clMDF_copy_float_array(__global float *input, int input_offset, __global float *output, int output_offset)
{
	
	unsigned int i = get_global_id(0);

	output[i + output_offset] = input[i + input_offset];
	
}

/**
 * Difference in response calculations, combined with inner product
 * Offsets are special cases
 * temp_e is a temporary buffer assumed to be at least of the same size as e
 */
__kernel void clMDF_combined_response_diff_inner_product(__global float *e, __global float *temp_e, __global float *y, __global float *input, 
						unsigned int C, 
						unsigned int window_size, 
						unsigned int frame_size,
						__global float *temp_odata,
						__global float *temp_e2,
						__local float *ldata)
{
	
	unsigned int i = get_global_id(0);
	unsigned int task = get_global_id(1);

	unsigned int tid = get_local_id(0);
	unsigned int xi;
	unsigned int yi;
	unsigned int n_inner_product = frame_size;

	barrier(CLK_GLOBAL_MEM_FENCE);
	if(task == 0) 
	{
		for(unsigned int chan=0;chan<C;chan++)
		{
			temp_e[chan*window_size + i] = e[chan*window_size+frame_size + i] - y[chan*window_size+frame_size + i];
			e[chan*window_size + i] = input[chan*frame_size + i] - y[chan*window_size+frame_size + i];
		}
		
		// Dbf
		// mdf_opencl_inner_prod(ocl_e, chan*st->window_size, ocl_e, chan*st->window_size, st->frame_size);
		// Runs on all threads, to make sure that they read/write to the same position 
		// in global memory as for the response diff.
		xi = i;
		yi = i;

		if(i < n_inner_product) {
			ldata[tid] = 0.0f;
			for(unsigned int chan = 0;chan<C;chan++)
				ldata[tid] += temp_e[xi + chan*window_size] * temp_e[yi + chan*window_size];
		}
		else
			ldata[tid] = 0.0f;

		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

		// do reduction in shared mem (accuracy problems on GPU?) 
		for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
		{
			if (tid < s)
			{
				ldata[tid] += ldata[tid + s];
				//ldata[tid] += 1.0f;
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

		// write result for this block to global mem
		if (tid == 0) 
		{
			temp_odata[get_group_id(0)] = ldata[0];
		}
	}
	else 
	{
		for(unsigned int chan=0;chan<C;chan++)
		{
			temp_e2[chan*window_size + i] = input[chan*frame_size + i] - y[chan*window_size+frame_size + i];
		}
		// See
		// mdf_opencl_inner_prod(ocl_e, chan*st->window_size, ocl_e, chan*st->window_size, st->frame_size);
		// Runs on all threads, to make sure that they read/write to the same position 
		// in global memory as for the response diff.
		xi = i;
		yi = i;

		if(i < n_inner_product) {
			ldata[tid] = 0.0f;
			for(unsigned int chan = 0;chan<C;chan++)
				ldata[tid] += temp_e2[xi + chan*window_size] * temp_e2[yi + chan*window_size];
		}
		else
			ldata[tid] = 0.0f;

		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

		// do reduction in shared mem (accuracy problems on GPU?) 
		for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
		{
			if (tid < s)
			{
				ldata[tid] += ldata[tid + s];
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

		// write result for this block to global mem
		if (tid == 0) 
		{
			temp_odata[get_group_id(0) + get_num_groups(0)] = ldata[0];
		}
	}
	
}

__kernel void clMDF_update_foreground_filter(__global float *e, __global float *window, __global float *y, unsigned int window_size, unsigned int frame_size)
{
	
	unsigned int i = get_global_id(0);
	unsigned int j = get_global_id(1);

	e[j*window_size+i+frame_size] = (window[i+frame_size] * e[j*window_size+i+frame_size]) + (window[i] * y[j*window_size+i+frame_size]);
	
} 

__kernel void clMDF_update_background_filter(__global float *y, __global float *e, __global float *input, unsigned int window_size, unsigned int frame_size)
{
	
	unsigned int i = get_global_id(0);
	unsigned int j = get_global_id(1);

	y[j*window_size+i+frame_size] = e[j*window_size+i+frame_size];
	e[j*window_size+i] = input[j*frame_size+i] - y[j*window_size+i+frame_size];
	
}

__kernel void clMDF_error_signal_combined(__global float *input, __global float *e, __global short *in, __global float *temp_results, __global float *temp_reductions, unsigned int window_size, unsigned int frame_size, unsigned int num_channels, __global int *saturated, __global float *y, __local float *sdata)
{
	unsigned int i = get_global_id(0);
	unsigned int j = get_global_id(1);
	unsigned int chan = get_global_id(1);
	unsigned int tid = get_local_id(0);
	unsigned int input_set = get_global_id(2);
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	// This is an arbitrary test for saturation in the microphone signal
	if (in[i*num_channels+chan] <= -32000 || in[i*num_channels+chan] >= 32000)
	{
		if (saturated[0] == 0)
			saturated[0] = 1;
	}
	
	// Sey 
	// clMDF_inner_product_reduce(ocl_e, chan*st->window_size+st->frame_size, ocl_y, chan*st->window_size+st->frame_size, g_odata, st->frame_size, sdata)  
	if(input_set == 1)
	{
		//for(int block_counter = 0;block_counter<(frame_size/get_local_size(0));block_counter++)
		temp_results[i*num_channels + chan] = input[chan*frame_size + i] - e[chan*window_size + frame_size + i];
		
		e[chan*window_size + frame_size + i] = e[chan*window_size + i];
		e[chan*window_size + i] = 0;
		
		barrier(CLK_GLOBAL_MEM_FENCE); 
		
		if(i < frame_size) {
			sdata[tid] = 0.0f;
			//for(chan = 0;chan<num_channels;chan++)
				sdata[tid] += e[chan*window_size + frame_size + i] * y[chan*window_size + frame_size + i];
				//sdata[tid] += 1.0f;
		}
		else
			sdata[tid] = 0.0f;

		barrier(CLK_LOCAL_MEM_FENCE);

		// do reduction in shared mem
		for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
		{
			if (tid < s)
			{
				sdata[tid] += sdata[tid + s];
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		// write result for this block to global mem
		if (tid == 0) 
		{
			temp_reductions[get_group_id(0)*num_channels + chan] = sdata[0];
			//temp_reductions[get_group_id(0) + chan] = 1.0f;
		}
		
		barrier(CLK_GLOBAL_MEM_FENCE);
	}

	// Syy 
	// clMDF_inner_product_reduce(ocl_y, chan*st->window_size+st->frame_size, ocl_y, chan*st->window_size+st->frame_size, g_odata, st->frame_size, sdata)
	else if(input_set == 2)
	{
		if(i < frame_size) {
			sdata[tid] = 0.0f;
			//for(chan = 0;chan<num_channels;chan++)
				//sdata[tid] += 1.0f;
				sdata[tid] += y[i + chan*window_size + frame_size] * y[i + chan*window_size + frame_size];
		}
		else
			sdata[tid] = 0;

		barrier(CLK_LOCAL_MEM_FENCE);

		// do reduction in shared mem
		for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
		{
			if (tid < s)
			{
				sdata[tid] += sdata[tid + s];
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		// write result for this block to global mem
		if (tid == 0) 
		{
			temp_reductions[get_num_groups(0)*num_channels + get_group_id(0)*num_channels + chan] = sdata[0];
			//temp_reductions[get_num_groups(0)*num_channels + get_group_id(0) + chan] = 512.0f;
		}
	}

	// Sdd 
	// clMDF_inner_product_reduce(ocl_input, chan*st->frame_size, ocl_input, chan*st->frame_size, g_odata, st->frame_size, sdata)
	else if(input_set == 3)
	{
		//unsigned int input_input1_offset = chan*frame_size;
		//unsigned int input_input2_offset = chan*frame_size;

		if(i < frame_size) {
			sdata[tid] = 0.0f;
			//for(chan = 0;chan<num_channels;chan++)
				sdata[tid] += input[i + chan*frame_size] * input[i + chan*frame_size];
				//sdata[tid] += 2.0f;
		}
		else
			sdata[tid] = 0;

		barrier(CLK_LOCAL_MEM_FENCE);

		// do reduction in shared mem
		for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
		{
			if (tid < s)
			{
				sdata[tid] += sdata[tid + s];
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		// write result for this block to global mem
		if (tid == 0) 
			//temp_reductions[get_num_groups(0)*num_channels*2 + get_group_id(0)*num_channels + chan] = 2048.0f;
			temp_reductions[get_num_groups(0)*num_channels*2 + get_group_id(0)*num_channels + chan] = sdata[0];
	}
}

__kernel void clMDF_smooth_far_end_compute_filtered_spectra(__global float *power, __global float *Xf, __global float *Rf, __global float *Eh, __global float *Yf, __global float *Yh,  float ss_1, float ss, float spec_average, unsigned int frame_size, __global float *temp_space, __global float *temp_eh_yh, __local float* sdata)
{
	unsigned int tid = get_global_id(0);
	unsigned int ltid = get_local_id(0);

	if(tid <= frame_size)
	{
		power[tid] = (ss_1 * power[tid]) + 1 + (ss * Xf[tid]);
	}
	
	// Calculate values to shared memory and update global memory values  
	if(tid <= frame_size)
	{
		//Eh_local = Rf[tid] - Eh[tid];
		//Yh_local = Yf[tid] - Yh[tid];
		sdata[ltid] = (Rf[tid] - Eh[tid]) * (Yf[tid] - Yh[tid]);
		sdata[get_local_size(0) + ltid] = (Yf[tid] - Yh[tid]) * (Yf[tid] - Yh[tid]);
		//sdata[ltid] = 1.0f;
		//sdata[get_local_size(0) + ltid] = 1.0f;
		Eh[tid] = (1-spec_average)*Eh[tid] + spec_average*Rf[tid];
		Yh[tid] = (1-spec_average)*Yh[tid] + spec_average*Yf[tid];
	}
	else
	{
		sdata[ltid] = 0.0f;
		sdata[get_local_size(0) + ltid] = 0.0f;
		//sdata[ltid] = 1.0f;
		//sdata[get_local_size(0) + ltid] = 1.0f;
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);

	// Do reduction in shared memory
	for(unsigned int s=get_local_size(0)/2; s>0; s>>=1) 
	{
		if (ltid < s)
		{
			sdata[ltid] += sdata[ltid + s];
			sdata[get_local_size(0) + ltid] += sdata[get_local_size(0) + ltid + s];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	// write result for this block to global mem
	if (ltid == 0) 
	{
		temp_space[2*get_group_id(0)] = sdata[0];	// Pey
		temp_space[2*get_group_id(0) + 1] = sdata[get_local_size(0)];	// Pyy
	}
}
 
__kernel void clMDF_learning_rate_calc(__global float *Yf, __global float *Rf, __global float *power_1, __global float *power, float RER, unsigned int frame_size, float leak_estimate, float adapt_rate, int adapted)
{
	unsigned int i = get_global_id(0);

	if(i < frame_size + 1)
	{
		if(adapted == 1)
		{
			float r, e;
			// Compute frequency-domain adaptation mask
			r = leak_estimate * Yf[i];
			e = Rf[i] + 1;
			if (r > 0.5f * e)
				r = 0.5f * e;
			r = (0.7f * r) + (0.3f * (float)(RER * e));
			power_1[i] = r / (e * power[i] + 10.0f);
		}
		else
		{
			power_1[i] = adapt_rate / (power[i] + 10.0f);
		}
	}
}

__kernel void clMDF_adapted_copy(__global float *last_y, __global short *in, __global int *out, unsigned int frame_size)
{
	unsigned int i = get_global_id(0);
	last_y[frame_size + i] = in[i] - out[i];
} 

/**
 *   Based on reduction kernel from the NVIDIA CUDA SDK (reduce2 kernel)
*/ 
__kernel void clMDF_inner_prod_power_spec_spectral_mul(__global float *Xf, __global float *data_time_domain, __global float *data_freq_domain, __global float *foreground, __global float *temp_reduction, __global float *acc, unsigned int N, unsigned int K, unsigned M, int frame_size, __local float* ldata)
{
	// mdf_opencl_power_spectrum_accum(data_freq_domain, 0, ocl_Xf, 0, st->window_size)
	if(get_global_id(2) == 0 && get_global_id(1) == 0)
	{
		unsigned int j = get_global_id(0);
		unsigned int i = j * 2 - 1;
		
		if(j == 0)
		{
			for(int speak=0;speak<K;speak++)
			{
				Xf[j] += data_freq_domain[speak*N + 0] * data_freq_domain[speak*N + 0];
				//Xf[j] = fma(data_freq_domain[speak*N + 0], data_freq_domain[speak*N + 0], Xf[j]);
				
				//Xf[j] += 1000.0f;
			}
			//Xf[j] = 9000.0f;
		}
		else if(i < (N-1))
		{
			for(int speak=0;speak<K;speak++)
			{
				Xf[j] += data_freq_domain[speak*N + i] * data_freq_domain[speak*N + i] + data_freq_domain[speak*N + i+1] * data_freq_domain[speak*N + i+1];
				//Xf[j] = fma(data_freq_domain[speak*N + i], data_freq_domain[speak*N + i], Xf[j]);
				//Xf[j] = fma(data_freq_domain[speak*N + i+1], data_freq_domain[speak*N + i+1], Xf[j]);
				//Xf[j] += 1000.0f;
			}
		}
		if(j == (get_global_size(0) - 1))
		{
			i += 2;
			for(int speak=0;speak<K;speak++)
			{
				Xf[j+1] += data_freq_domain[speak*N + i] * data_freq_domain[speak*N + i]; 
				//Xf[j+1] = fma(data_freq_domain[speak*N + i], data_freq_domain[speak*N + i], Xf[j+1]);
				//Xf[j + 1] += 1000.0f;
			}
		} 
	}
	// mdf_opencl_spectral_mul_accum(data_freq_domain, ocl_foreground, 0, ocl_Y, 0, st->window_size, st->M*st->K)
	// Only first part, needs reduce kernel as well
	// X=data_freq_domain
	// Y=foreground
	else if(get_global_id(2) == 1)
	{
		int i = get_global_id(0);
		unsigned int tid = get_local_id(0);
		unsigned int chan = get_global_id(1);
		unsigned int c;
		unsigned int M_org = M / K;
		
		// Temporary accum registers
		float tmp1 = 0.0f;
		float tmp2 = 0.0f;
		float tmp3 = 0.0f;

		if(i == 0)
		{
			for(c=0;c<M;c++)
			{
				tmp1 += data_freq_domain[c*N + 0]*foreground[chan*N*K*M_org + c*N + 0];
				tmp2 += (data_freq_domain[c*N + 1]*foreground[chan*N*K*M_org + c*N + 1] - data_freq_domain[c*N + 2]*foreground[chan*N*K*M_org + c*N + 2]);
				//tmp1 += 1.0f;
				//tmp2 += 1.0f;
			}
		}
		else if(i < (get_global_size(0) - 1))
		{
			for(c=0;c<M;c++)
			{
				tmp1 += (data_freq_domain[c*N + 2*i - 1]*foreground[chan*N*K*M_org + c*N + 2*i - 1] - data_freq_domain[c*N + 2*i]*foreground[chan*N*K*M_org + c*N + 2*i]);
				tmp2 += (data_freq_domain[c*N + 2*i]*foreground[chan*N*K*M_org + c*N + 2*i - 1] + data_freq_domain[c*N + 2*i - 1]*foreground[chan*N*K*M_org + c*N + 2*i]);
				//tmp1 += 2.0f;
				//tmp2 += 2.0f; 
			}
		}
		else if(i == (get_global_size(0) - 1))
		{	
			for(c=0;c<M;c++)
			{
				tmp1 += (data_freq_domain[c*N + 2*i - 1]*foreground[chan*N*K*M_org + c*N + 2*i - 1] - data_freq_domain[c*N + 2*i]*foreground[chan*N*K*M_org + c*N + 2*i]);
				tmp2 += (data_freq_domain[c*N + 2*i]*foreground[chan*N*K*M_org + c*N + 2*i - 1] + data_freq_domain[c*N + 2*i - 1]*foreground[chan*N*K*M_org + c*N + 2*i]);
				tmp3 += data_freq_domain[c*N + (N-1)]*foreground[chan*N*K*M_org + c*N + (N-1)];
				//tmp1 += 3.0f;
				//tmp2 += 3.0f;
			}
		}
		
		// Write out to global memory
		if(i > 0 && i < (get_global_size(0) - 1))
		{
			acc[chan*N + 2*i - 1] = tmp1;
			acc[chan*N + 2*i] = tmp2;
			//acc[chan*N + 2*i - 1] = 9000.0f;
			//acc[chan*N + 2*i] = 8000.0f;
		}
		else if(i == 0)
		{
			acc[chan*N] = tmp1;
			//acc[chan*N + 2*i + 1] = tmp2;
		}
		else if(i == (get_global_size(0) - 1))
		{
			acc[chan*N + 2*i - 1] = tmp1;
			acc[chan*N + 2*i] = tmp2;
			acc[chan*N + 2*i + 1] = tmp3;
		}
	}
}
