#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <string.h>
#include <fcntl.h>
#include "speex/speex_echo.h"
#include "speex/speex_preprocess.h"

#define NN 128
#define TAIL 128*20

#define MAX_ITERATIONS 20
// If benchmarking on number of channels, set these values to the maximum value that
// is going to be tested
#define NUM_CHANNELS 1	// OpenCL version is limited to max 64 channels
#define NUM_SPEAKERS 1
//#define ENABLE_LOAD_BAL 1

// Uncomment for benchmarking increasing tail lengths. DO NOT COMBINE with other benchmarks!
//#define BENCHMARK_TAIL_LENGTH 1
//#define BENCHMARK_TAIL_LENGTH_MIN_LENGTH 1024
//#define BENCHMARK_TAIL_LENGTH_MAX_LENGTH 20480
//#define BENCHMARK_TAIL_LENGTH_STEP_SIZE 1024

// Uncomment for benchmarking increasing channel number
//#define BENCHMARK_INPUT_CHANS 1
//#define BENCHMARK_INPUT_CHANS_MIN_CHANS 1
//#define BENCHMARK_INPUT_CHANS_MAX_CHANS 16
//#define BENCHMARK_INPUT_CHANS_STEP_SIZE 1

// Uncomment for benchmarking increasing channel number, can be combined with input channels
//#define BENCHMARK_OUTPUT_CHANS 1
//#define BENCHMARK_OUTPUT_CHANS_MIN_CHANS 1
//#define BENCHMARK_OUTPUT_CHANS_MAX_CHANS 16
//#define BENCHMARK_OUTPUT_CHANS_STEP_SIZE 1

enum MDF_DEVICE            /* Defines an enumeration type    */
{
    OPENCL = 0,
    CPU_REFERENCE = 1,
};

// Load monitor global variables
FILE *fp;
char *cpustring;
struct timeval last_cpu_check, curr_time;
int exec_dev = CPU_REFERENCE;
int req_dev = CPU_REFERENCE;
int prev_total = 0;
int prev_idle = 0;
	
int get_system_CPU_load();
void interleave_input(short *echo_in, short *ref_in, short *echo_out, short *ref_out);

int main(int argc, char **argv)
{
   FILE *echo_fd[NUM_SPEAKERS];
   FILE *ref_fd[NUM_CHANNELS];
   //FILE *echo_fd1, *echo_fd2;
   //FILE *ref_fd1, *ref_fd2;
   FILE *e_fd[2];	// Write one output from CPU and one from OpenCL
   FILE *errors_file_fd;
   // Frame read from file
   short *echo_buf = (short *)malloc(sizeof(short)*NN*NUM_SPEAKERS);	
   // Frame passed to filter, interleaved channels
   short *echo_buf_interleaved = (short *)malloc(sizeof(short)*NN*NUM_SPEAKERS);	
   short *ref_buf = (short *)malloc(sizeof(short)*NN*NUM_CHANNELS);
   short *ref_buf_interleaved = (short *)malloc(sizeof(short)*NN*NUM_CHANNELS);
   short *e_buf = (short *)malloc(sizeof(short)*NN*NUM_CHANNELS);
   short *ref_buf_copy = (short *)malloc(sizeof(short)*NN*NUM_CHANNELS);
   short *echo_buf_copy = (short *)malloc(sizeof(short)*NN*NUM_SPEAKERS);
   short *e_buf_copy = (short *)malloc(sizeof(short)*NN*NUM_CHANNELS);
   float *residual_error = (float *)malloc(sizeof(double)*MAX_ITERATIONS);
   SpeexEchoState *st;
   SpeexEchoState *st_copy;
   SpeexPreprocessState *den;
   SpeexPreprocessState *den_copy;
   int sampleRate = 48000;
   int last_cpu_usage = 0;
   int i = 0;
   int j = 0;
   int run = 1;
   char clear_key;
   
   cpustring = (char *)malloc(sizeof(char) * 80);
   #ifdef ENABLE_LOAD_BAL
   gettimeofday(&last_cpu_check, NULL);
   #endif
   
   speex_echo_opencl_init_dev(NN);
   
   #if defined BENCHMARK_TAIL_LENGTH
   int t = 0;
   for(t=BENCHMARK_TAIL_LENGTH_MIN_LENGTH;t<=BENCHMARK_TAIL_LENGTH_MAX_LENGTH;t+=BENCHMARK_TAIL_LENGTH_STEP_SIZE)
   {
   #endif
   #if defined BENCHMARK_INPUT_CHANS
   int c = 0;
   for(c=BENCHMARK_INPUT_CHANS_MIN_CHANS;c<=BENCHMARK_INPUT_CHANS_MAX_CHANS;c+=BENCHMARK_INPUT_CHANS_STEP_SIZE)
   {
   #endif
   #if defined BENCHMARK_OUTPUT_CHANS
   int k = 0;
   for(k=BENCHMARK_OUTPUT_CHANS_MIN_CHANS;k<=BENCHMARK_OUTPUT_CHANS_MAX_CHANS;k+=BENCHMARK_OUTPUT_CHANS_STEP_SIZE)
   {
   #endif
   for(i=0;i<MAX_ITERATIONS;i++)
      residual_error[i] = 0.0;
   
   if (argc != 4)
   {
      fprintf(stderr, "testecho mic_signal.sw speaker_signal.sw output.sw\n");
      exit(1);
   }
   // Output input files, just open several version of the single-channel files
   for(i=0;i<NUM_CHANNELS;i++)
   {
      printf("Open loop input channels.\n");
      ref_fd[i] = fopen(argv[1], "rb");
   }
   for(i=0;i<NUM_SPEAKERS;i++)
   {
      printf("Open loop output channels.\n");
      echo_fd[i] = fopen(argv[2], "rb");
   }
   // Write to two output files
   printf("Opening regular output file.\n");
   e_fd[0] = fopen(argv[3], "wb");
   printf("Opening OpenCL output file.\n");
   char *ocl_out_string = (char *)malloc(sizeof(char)*500);
   sprintf(ocl_out_string, "%s_ocl", argv[3]);
   e_fd[1] = fopen(ocl_out_string, "wb");

   printf("Tried to open files.\n");
   
   // Check that the output file could be opened
   if(e_fd[0] == NULL)
   {
      fprintf(stderr, "The output file could not be opened for writing!\n");
      exit(1);  
   }
   else if(e_fd[1] == NULL)
   {
      fprintf(stderr, "The output file for OpenCL could not be opened for writing!\n");
      exit(1);  
   }
   // Check that the input files for all channels could be opened
   else
   {
      for(i=0;i<NUM_SPEAKERS;i++)
      {
         if(echo_fd[i] == NULL)
	 {
	    fprintf(stderr, "One of the given filenames does not exist!\n");
	    exit(1); 
	 }
      }
      for(i=0;i<NUM_CHANNELS;i++)
      {
         if(ref_fd[i] == NULL)
	 {
	    fprintf(stderr, "One of the given filenames does not exist!\n");
	    exit(1); 
	 }
      }
   }
   printf("Files are open.\n");
   
   printf("Running regular echo cancellation.\n");
   #if defined BENCHMARK_TAIL_LENGTH
   st = speex_echo_state_init_mc(NN, t, NUM_CHANNELS, NUM_SPEAKERS);
   st_copy = speex_echo_state_init_mc(NN, t, NUM_CHANNELS, NUM_SPEAKERS);
   #elif defined BENCHMARK_OUTPUT_CHANS && defined BENCHMARK_INPUT_CHANS
   st = speex_echo_state_init_mc(NN, TAIL, c, k);
   st_copy = speex_echo_state_init_mc(NN, TAIL, c, k);
   #elif defined BENCHMARK_INPUT_CHANS
   st = speex_echo_state_init_mc(NN, TAIL, c, NUM_SPEAKERS);
   st_copy = speex_echo_state_init_mc(NN, TAIL, c, NUM_SPEAKERS);
   #elif defined BENCHMARK_OUTPUT_CHANS
   st = speex_echo_state_init_mc(NN, TAIL, NUM_CHANNELS, k);
   st_copy = speex_echo_state_init_mc(NN, TAIL, NUM_CHANNELS, k);
   #else
   st = speex_echo_state_init_mc(NN, TAIL, NUM_CHANNELS, NUM_SPEAKERS);
   st_copy = speex_echo_state_init_mc(NN, TAIL, NUM_CHANNELS, NUM_SPEAKERS);
   #endif
   den = speex_preprocess_state_init(NN, sampleRate);
   den_copy = speex_preprocess_state_init(NN, sampleRate);
   speex_echo_ctl(st, SPEEX_ECHO_SET_SAMPLING_RATE, &sampleRate);
   speex_echo_ctl(st_copy, SPEEX_ECHO_SET_SAMPLING_RATE, &sampleRate);
   speex_preprocess_ctl(den, SPEEX_PREPROCESS_SET_ECHO_STATE, st);
   speex_preprocess_ctl(den_copy, SPEEX_PREPROCESS_SET_ECHO_STATE, st_copy);
   
   speex_echo_opencl_init_mem(st);
   speex_echo_opencl_execution_time_init();

   //while (run)
   while (run && i < MAX_ITERATIONS)
   //while (run && i < 200)
   {
      //printf("Starting loop iteration echo.\n");
      // Check if end of file has been reached, for all channels
      for(j=0;j<NUM_SPEAKERS;j++)
      {
	  if(feof(echo_fd[j]))
		  run = 0;
      }
      for(j=0;j<NUM_CHANNELS;j++)
      {
	  if(feof(ref_fd[j]))
		  run = 0;
      }
      if(run)
      {
	 for(j=0;j<NUM_SPEAKERS;j++)
	 {
	    fread(echo_buf + j*NN, sizeof(short), NN, echo_fd[j]); 
	 }
	 for(j=0;j<NUM_CHANNELS;j++)
	 {
	    fread(ref_buf + j*NN, sizeof(short), NN, ref_fd[j]); 
	 }
	 interleave_input(echo_buf, ref_buf, echo_buf_interleaved, ref_buf_interleaved);
         init_checkpoints();
         init_timingpoints();
	 // Leave commented if OpenCL state sync is used for each frame!
         //if(i == 0)
	 	copy_echo_state(st, st_copy);
         memcpy(ref_buf_copy, ref_buf_interleaved, sizeof(short)*NN*NUM_CHANNELS);
         memcpy(echo_buf_copy, echo_buf_interleaved, sizeof(short)*NN*NUM_SPEAKERS);
         //memcpy(e_buf_copy, e_buf, sizeof(short)*NN*NUM_CHANNELS);
	 #ifdef ENABLE_LOAD_BAL
	 if(exec_dev == CPU_REFERENCE)
	 #endif
		speex_echo_cancellation(st, ref_buf_interleaved, echo_buf_interleaved, e_buf);
	 #ifdef ENABLE_LOAD_BAL
	 else if(exec_dev == OPENCL)
		 speex_echo_cancellation_opencl(st_copy, ref_buf_copy, echo_buf_copy, e_buf);
	 #else
		speex_echo_cancellation_opencl(st_copy, ref_buf_copy, echo_buf_copy, e_buf_copy);
	 #endif
         cleanup_checkpoints();
         cleanup_timingpoints();
	 // Store residual error divided by number of frames
	 if(i<MAX_ITERATIONS)
		residual_error[i] = get_residual_error(st, st_copy) / NN;
         
	 // Set to e_buf_copy to output OpenCL output to file
         speex_preprocess_run(den, e_buf);
	 fwrite(e_buf, sizeof(short), NN*NUM_CHANNELS, e_fd[0]);
	 #ifndef ENABLE_LOAD_BAL
	 speex_preprocess_run(den_copy, e_buf_copy);
	 fwrite(e_buf_copy, sizeof(short), NN*NUM_CHANNELS, e_fd[1]);
	 #endif
         
         i++;
	 #ifdef ENABLE_LOAD_BAL
	 gettimeofday(&curr_time, NULL);
         // Every second, check CPU usage
         if(((long long) (curr_time.tv_sec - last_cpu_check.tv_sec) * 1000000 + (curr_time.tv_usec - last_cpu_check.tv_usec)) > 1000000.0)
         {
            last_cpu_usage = get_system_CPU_load();
	    last_cpu_check = curr_time;
	    
	    if(last_cpu_usage > 50)
	    {
		    if(exec_dev == CPU_REFERENCE)
		    {
			printf("Switching to OpenCL exeuction (i=%d, cpu=%d).\n", i, last_cpu_usage);
			exec_dev = OPENCL;
			copy_echo_state(st, st_copy);
			speex_echo_opencl_sync_to_GPU(st);
		    }
	    }
	    else
	    {
		    if(exec_dev == OPENCL)
		    {
			printf("Switching to reference CPU exeuction (i=%d, cpu=%d).\n", i, last_cpu_usage);
			exec_dev = CPU_REFERENCE;
			printf("Copying echo state\n");
			copy_echo_state(st_copy, st);
			printf("Syncing buffers\n");
			speex_echo_opencl_sync_to_CPU(st);
		    }
	    }
	    
	    printf("CPU use: %d\n", last_cpu_usage);
         }
         #endif
      }
      //printf("Ended loop iteration echo.\n");
   }

   // Write out residual errors to file
   errors_file_fd = fopen("errors.txt", "wb");
   if(errors_file_fd != NULL)
   {
      for(i=0;i<MAX_ITERATIONS;i++)
	  fprintf(errors_file_fd, "%f ", residual_error[i]);
   }
   fclose(errors_file_fd);
	
   speex_echo_opencl_print_execution_time_stats();
   printf("Cleaning up execution time.\n");
   speex_echo_opencl_execution_time_cleanup();

   printf("Cleaning up original state.\n");
   speex_echo_state_destroy(st);
   printf("Cleaning up copy state.\n");
   speex_echo_state_destroy(st_copy);
   printf("Cleaning up OpenCL.\n");
   speex_echo_opencl_cleanup_mem();

   speex_preprocess_state_destroy(den);
   speex_preprocess_state_destroy(den_copy);
   for(i=0;i<NUM_SPEAKERS;i++)
   {
      fclose(echo_fd[i]);
   }
   for(i=0;i<NUM_CHANNELS;i++)
   {
      fclose(ref_fd[i]);
   }
   for(i=0;i<2;i++)
   {
      fclose(e_fd[i]);
   }
   
   #if defined BENCHMARK_TAIL_LENGTH
      printf("Done with benchmark tail length %d, press a key to continue.\n", t);
      scanf("%c", &clear_key);
   }
   #endif
   #if defined BENCHMARK_OUTPUT_CHANS
      printf("Done with benchmark %d output channels, press a key to continue.\n", k);
      scanf("%c", &clear_key);
   }
   #endif
   #if defined BENCHMARK_INPUT_CHANS
      printf("Done with benchmark %d input channels, press a key to continue.\n", c);
      scanf("%c", &clear_key);
   }
   #endif
   speex_echo_opencl_cleanup_dev();
   
   printf("Cleaning up copy arrays.\n");
   free(residual_error);
   free(ref_buf_copy);
   free(echo_buf_copy);
   free(e_buf_copy);
   free(echo_buf);
   free(echo_buf_interleaved);
   free(ref_buf);
   free(ref_buf_interleaved);
   free(e_buf);
   printf("Return, exiting.\n");
   return 0;
}

/**
 * Linux-specific function for retrieving system CPU load
 */
int get_system_CPU_load()
{		
	int diff_idle;
	int diff_total;
	int diff_usage;
	int total;
	int idle;
	
	int stat_user = 0;
	int stat_nice = 0;
	int stat_system = 0;
	int stat_idle = 0;
	int stat_iowait = 0;
	int stat_irq = 0;
	int stat_softirq = 0;
	
	// Very simple/naive cpu load monitor for Linux systems
	// Source: http://colby.id.au/node/39
	if((fp = fopen("/proc/stat", "r")) != NULL) {
		
		fgets(cpustring, 80, fp);
		//printf("/proc/stat:\n%s", cpustring);
		sscanf(cpustring, "cpu  %d %d %d %d %d %d %d", &stat_user, &stat_nice, &stat_system, &stat_idle, &stat_iowait, &stat_irq, &stat_softirq);

		total = stat_user + stat_nice + stat_system + stat_idle + stat_iowait + stat_irq + stat_softirq;

		idle = stat_idle;

		diff_idle = idle - prev_idle;
		diff_total = total - prev_total;
		if(diff_total > 0)
		{
			//printf("Calc: (1000*(%d-%d)/%d + 5) / 10\n", diff_total, diff_idle, diff_total);
			diff_usage = (1000*(diff_total-diff_idle)/diff_total + 5) / 10;
		}

		prev_total = total;
		prev_idle = idle;

		fclose(fp);
		
		return diff_usage;
	}
	else
		return -1;
	
}

/**
 * Utility-function to interleave multichannel input from files
 * _in buffers contain concecutive frames from each of the channels:
 * 	1. frame chan 1, 1. frame chan 2, 2. frame chan 1, 2. frame chan 2 etc...
 * _out buffers contains the _in buffers interleaved:
 * 	1. sample chan 1, 1. sample chan 2, 2.sample chan 1, 2.sample chan 2 etc...
 */
void interleave_input(short *echo_in, short *ref_in, short *echo_out, short *ref_out)
{
	int i, speak, chan;
	for(speak=0;speak<NUM_SPEAKERS;speak++)
	{
		for(i=0;i<NN;i++)
		{
			echo_out[i*NUM_SPEAKERS + speak] = echo_in[speak*NN + i];
		}
	}
	for(chan=0;chan<NUM_CHANNELS;chan++)
	{
		for(i=0;i<NN;i++)
		{
			ref_out[i*NUM_CHANNELS + chan] = ref_in[chan*NN + i];
		}
	}
}
